#!/usr/bin/env python
# coding=utf-8

import re
import typing
from collections import defaultdict
from functools import partial, reduce
from itertools import chain, count, product
from operator import and_, mul
from queue import Queue

import cloudpickle
import attr
from cached_property import cached_property
import numpy as np
from loguru import logger
from more_itertools import unique_everseen
from sympy import (
    Derivative,
    Eq,
    Function,
    Indexed,
    KroneckerDelta,
    Number,
    Symbol,
    Wild,
    Dummy,
    oo,
    solve,
    sympify,
)

from .spatial_schemes import FiniteDifferenceScheme, chain_schemes, upwind
from .variables import Coordinate, Unknown, _convert_coord_list, _convert_unk_list


def _convert_pde_list(pdes):
    if isinstance(pdes, str):
        return [pdes]
    else:
        return pdes


def _partial_derivative(expr, symbolic_coordinates):
    # proxy function that can be easily curried (with functools.partial)
    return Derivative(expr, *symbolic_coordinates)


def _build_sympy_namespace(equation, coordinates, unknowns, parameters):
    """ Check the equation, find all the derivative in Euler notation
    (see https://en.wikipedia.org/wiki/Notation_for_differentiation#Euler's_notation)
    the way that dxxU will be equal to Derivative(U(x), x, x).
    All the derivative found that way are add to a subsitution rule as dict and
    applied when the equation is sympified.
    """

    # look at all the dxxU, dxyV... and dx(...), dxy(...) and so on in the equation
    spatial_derivative_re = re.compile(
        r"d(?P<derargs>\w+?)(?:(?P<depder>(?:%s)+)|\((?P<inder>.*?)\))"
        % "|".join([var.name for var in chain(unknowns, parameters)])
    )
    spatial_derivatives = spatial_derivative_re.findall(str(equation))

    # they can be derivatives inside the dx(...), we check it until there is no more
    queue = Queue()
    [queue.put(sder[2]) for sder in spatial_derivatives if sder[2]]
    while not queue.empty():
        inside_derivative = queue.get()
        new_derivatives = spatial_derivative_re.findall(inside_derivative)
        [queue.put(sder[2]) for sder in new_derivatives if sder[2]]
        spatial_derivatives.extend(new_derivatives)
    # The sympy namespace is built with...
    namespace = dict()
    # All the coordinates
    namespace.update({coord.name: coord.symbol for coord in coordinates})
    # All the dependent variables: unk and parameters
    namespace.update(
        {
            dvar.name: (
                dvar.symbol(*[coord.symbol for coord in dvar.coordinates])
                if dvar.coordinates
                else dvar.symbol
            )
            for dvar in chain(unknowns, parameters)
        }
    )
    # All the harversted derivatives
    for coord, unk, _ in spatial_derivatives:
        if unk:
            namespace["d%s%s" % (coord, unk)] = _partial_derivative(
                namespace[unk], coord
            )
        else:
            namespace["d%s" % coord] = partial(
                _partial_derivative, symbolic_coordinates=coord
            )
    return namespace


def _build_keep_derivs_namespace(equation):
    """ Check the equation, find all the derivative in Euler notation
    (see https://en.wikipedia.org/wiki/Notation_for_differentiation#Euler's_notation)
    the way that dxxU will be equal to Derivative(U(x), x, x).
    All the derivative found that way are add to a subsitution rule as dict and
    applied when the equation is sympified.
    """

    # look at all the dxxU, dxyV... and dx(...), dxy(...) and so on in the equation
    spatial_derivative_re = re.compile(r"D(?P<derargs>\w+?)(?:\((?P<inder>.*)\))")
    spatial_derivatives = spatial_derivative_re.findall(str(equation))

    return spatial_derivatives


def list_node_coords(domain):
    available_node_coords = {}
    for coord, (left_cond, right_cond) in domain.items():
        node_coords = []
        node_coords.append(coord.idx)
        # for both side, left|right_cond can be true : in that case, no coords has to
        # be added : all that side is in the bulk.
        try:
            node_coords.extend(np.arange(coord.idx.lower, left_cond.rhs))
        except AttributeError:
            pass
        try:
            node_coords.extend(
                np.arange(right_cond.rhs - coord.N + 1, coord.idx.upper - coord.N + 1)
                + coord.N
            )
        except AttributeError:
            pass
        available_node_coords[coord] = node_coords
    return available_node_coords


def list_conditions(domain):
    available_conds = {}
    for coord, (left_cond, right_cond) in domain.items():
        conds = []
        conds.append(left_cond & right_cond)
        # for both side, left|right_cond can be true : in that case, no coords has to
        # be added : all that side is in the bulk.
        try:
            conds.extend(
                [
                    Eq(coord.idx, node_coord)
                    for node_coord in np.arange(coord.idx.lower, left_cond.rhs)
                ]
            )
        except AttributeError:
            pass
        try:
            conds.extend(
                [
                    Eq(coord.idx, node_coord)
                    for node_coord in np.arange(
                        right_cond.rhs - coord.N + 1, coord.idx.upper - coord.N + 1
                    )
                    + coord.N
                ]
            )
        except AttributeError:
            pass
        available_conds[coord] = conds
    return available_conds


@attr.s
class PDEquation:
    equation = attr.ib(type=str)
    unknowns = attr.ib(type=typing.Sequence[Unknown], converter=_convert_unk_list)
    parameters = attr.ib(
        type=typing.Sequence[Unknown], converter=_convert_unk_list, default=[]
    )
    subs = attr.ib(type=dict, default={})
    boundary_conditions = attr.ib(type=dict, factory=dict)
    schemes = attr.ib(
        type=typing.Sequence[FiniteDifferenceScheme],
        default=(FiniteDifferenceScheme(),),
        repr=False,
    )
    symbolic_equation = attr.ib(init=False, repr=False)
    fdiff = attr.ib(init=False, repr=False)
    raw = attr.ib(type=bool, default=False, repr=False)
    dirichlet_nodes = attr.ib(
        type=typing.Sequence[typing.Tuple[int, ...]], factory=list, repr=False
    )

    def __attrs_post_init__(self):
        self._t = Symbol("t")
        self._complete_coordinates()
        if self.raw:
            # For "raw" equations already in discretized form as periodic bc
            self.fdiff = sympify(
                self.equation, locals={unk.name: unk.discrete for unk in self.unknowns}
            )
            return
        self._fill_incomplete_unknowns()

        self._sympy_namespace = _build_sympy_namespace(
            self.equation, self.coordinates, self.unknowns, self.parameters
        )
        self._sympify_equation()

        # substitute the auxiliary definition
        sympified_subs = {
            sympify(aux_key, locals=self._sympy_namespace): sympify(
                aux_value, locals=self._sympy_namespace
            )
            for aux_key, aux_value in self.subs.items()
        }
        self.symbolic_equation = self.symbolic_equation.subs(sympified_subs)

        self._as_finite_diff()
        self._reduce_coordinates()

    def __getstate__(self):
        return {key: cloudpickle.dumps(value) for key, value in self.__dict__.items()}

    def __setstate__(self, state):
        self.__dict__.update(
            {key: cloudpickle.loads(value) for key, value in state.items()}
        )

    @property
    def parsed_boundary_conditions(self):
        return dict(self._parse_bc(self.boundary_conditions))

    def _reduce_coordinates(self):
        variables = [
            self.mapper[str(indexed.base)] for indexed in self.fdiff.atoms(Indexed)
        ]
        available_unks = [
            variable for variable in variables if isinstance(variable, Unknown)
        ]
        real_coordinates = set().union(*[unk.coordinates for unk in available_unks])
        self.coordinates = sorted(real_coordinates)

    def _non_expendable_deriv(self, wrts, arg):
        arg = sympify(arg)
        dummy = Function(str(Dummy()))
        dummy_deriv = Derivative(dummy(*self.coordinates), wrts)
        fdiff_dummy = chain_schemes(self.schemes, dummy_deriv)

        def replace_dummy(*coords):
            return arg.subs(
                {
                    symbolic_coord.name: coord
                    for symbolic_coord, coord in zip(self.coordinates, coords)
                }
            )

        return fdiff_dummy.replace(dummy, replace_dummy)

    @property
    def stencils(self):
        stencils = {}
        ref_pde = self([0] * len(self.coordinates))
        indexed = ref_pde.atoms(Indexed)
        for unk in self.unknowns:
            coords = set(
                [
                    tuple(map(int, index.indices))
                    for index in indexed
                    if unk.discrete == index.args[0]
                ]
            )
            ptp = np.array(list(coords)).ptp() + 1
            center = ptp // 2
            stencil_array = np.zeros([ptp] * len(unk.coords), dtype=bool)
            idxs = [np.array(coord + center, dtype=int) for coord in zip(*coords)]
            stencil_array[tuple(idxs)] = True
            stencils[unk.name] = stencil_array
        return stencils

    def _fill_incomplete_unknowns(self):
        """fill every dependent variables that lack information on
        independent variables with the global independent variables
        """
        for i, unk in enumerate(self.unknowns):
            if not unk.coordinates:
                object.__setattr__(self.unknowns[i], "coordinates", self.coordinates)

    def _complete_coordinates(self):
        """if independent variables is not set, extract them from
        the dependent variables. If not set in dependent variables,
        1D resolution with "x" as independent variable is assumed.
        """
        harvested_coords = list(
            chain(
                *[
                    dep_var.coordinates
                    for dep_var in self.unknowns
                    if dep_var.coordinates is not None
                ]
            )
        )
        if harvested_coords:
            self.coordinates = harvested_coords
        else:
            self.coordinates = _convert_coord_list(["x"])
        self.coordinates = list(unique_everseen(self.coordinates))

    def _sympify_equation(self):
        self.symbolic_equation = sympify(self.equation, locals=self._sympy_namespace)
        for unk in self.unknowns:
            self.symbolic_equation = self.symbolic_equation.subs(unk.name, unk.symbol)
        self.symbolic_equation = self.symbolic_equation.doit()

    def _as_finite_diff(self):

        fdiff = self.symbolic_equation
        fdiff = chain_schemes(self.schemes, fdiff)
        fdiff = fdiff.replace(Function("upwind"), upwind)
        to_keep_derivs = _build_keep_derivs_namespace(fdiff)
        for wrts, arg in to_keep_derivs:
            fdiff = fdiff.replace(
                Function("D%s" % "".join(wrts)),
                partial(self._non_expendable_deriv, wrts),
            )

        for coord in self.coordinates:
            a = Wild("a", exclude=[coord.step, coord.symbol, 0])
            fdiff = fdiff.replace(coord.symbol + a * coord.step, coord.idx + a)

        for var in chain(self.unknowns, self.parameters):

            def replacement(*args):
                return var.discrete[args]

            if var.coordinates:
                fdiff = fdiff.replace(var.symbol, replacement)

        for indexed in fdiff.atoms(Indexed):
            new_indexed = indexed.subs(
                {coord.symbol: coord.idx for coord in self.coordinates}
            )
            fdiff = fdiff.subs(indexed, new_indexed)

        fdiff = fdiff.subs(
            {coord.symbol: coord.discrete[coord.idx] for coord in self.coordinates}
        )

        self.fdiff = fdiff

    def __call__(self, coordinates):
        logger.trace("evaluate pde %s at %s" % (self.equation, coordinates))
        subs = {
            coord.idx: coord_value
            for coord, coord_value in zip(self.coordinates, coordinates)
        }
        logger.trace("subs: %s" % subs)
        return self.fdiff.subs(subs)

    def __str__(self):
        return self.__repr__()

    def domains(self, *node_coords):
        return tuple(
            [
                coord.domain(node_coord)
                for coord, node_coord in zip(self.coordinates, node_coords)
            ]
        )

    @property
    def unknowns_dict(self):
        return {unk.name: unk for unk in self.unknowns}

    @property
    def coordinates_dict(self):
        return {
            coord.name: coord
            for coord in set(chain(*[unk.coords for unk in self.unknowns]))
        }

    @property
    def parameters_dict(self):
        return {par.name: par for par in self.parameters}

    @property
    def mapper(self):
        return dict(
            **self.unknowns_dict, **self.parameters_dict, **self.coordinates_dict
        )

    @property
    def physical_domains(self):
        domains = product(*[("left", "bulk", "right") for coord in self.coordinates])
        conds = product(
            *[
                (
                    Eq(coord.idx, coord.idx.lower),
                    (coord.idx.lower < coord.idx) & (coord.idx < coord.idx.upper),
                    Eq(coord.idx, coord.idx.upper),
                )
                for coord in self.coordinates
            ]
        )
        return dict(zip(domains, [reduce(and_, cond) for cond in conds]))

    @property
    def node_coords(self):
        domains = product(*[("left", "bulk", "right") for coord in self.coordinates])

        return [
            tuple(
                coord.get_node_coord(domain)
                for coord, domain in zip(self.coordinates, domain)
            )
            for domain in domains
        ]

    @property
    def node_coords_subs(self):

        return [
            {
                coord.idx: node_coord
                for coord, node_coord in zip(self.coordinates, node_coords)
            }
            for node_coords in self.node_coords
        ]

    def get_domains(self, indexed):
        unk = self.mapper[str(indexed.base)]
        return unk.domains(*indexed.indices)

    def get_distances(self, indexed):
        unk = self.mapper[str(indexed.base)]
        return unk.distances(*indexed.indices)

    def is_in_bulk(self, indexed):
        return not self.is_outside(indexed)

    def is_outside(self, indexed):
        return any(self.get_distances(indexed))

    def _distance_node_to_other(self, node, other_node):
        return tuple(np.abs(np.array(node) - np.array(other_node)))

    def _walk_in_domain(self, node, domains):
        next_node = tuple(node)
        for distance in count():
            local_pde = self(next_node)
            outside_indexed = list(filter(self.is_outside, local_pde.atoms(Indexed)))
            if not outside_indexed:
                break
            next_node = [
                idx + sign_distance(1, domain)
                for idx, domain in zip(next_node, domains)
            ]
        product_distances = product(*[range(distance)] * len(node))
        for distances in product_distances:
            yield tuple(
                [
                    idx + sign_distance(distance, domain)
                    for idx, domain, distance in zip(node, domains, distances)
                ]
            )

    def _parse_bc(self, bcs):
        if isinstance(bcs, str):
            default_bcs = defaultdict(lambda: bcs)
        else:
            default_bcs = defaultdict(lambda: "noflux", bcs)
        unks = [*self.unknowns, *self.parameters]
        for unk, axis in chain(*[product([unk], unk.coords) for unk in unks]):
            yield (unk, axis), BoundaryCondition(
                unk, axis, self, default_bcs[(unk.name, axis.name)]
            )

    # long running, should not be in property
    @cached_property
    def computation_nodes(self):
        extended_computation_nodes = []
        for new_node in chain(
            *[
                self._walk_in_domain(node, domains)
                for node, domains in zip(self.node_coords, self.physical_domains)
            ]
        ):
            extended_computation_nodes.append(new_node)
        return list(set(extended_computation_nodes).union(self.node_coords))

    def _get_filtered_cross(self, coord, node_coords):
        coord_idx = self.coordinates.index(coord)
        cross_node_coords = tuple(
            [node for i, node in enumerate(node_coords) if i != coord_idx]
        )
        cross_computation_nodes = [
            tuple([node for i, node in enumerate(nodes) if i != coord_idx])
            for nodes in self.computation_nodes
        ]
        for i, (computation_node_coords, cross_computation_nodes) in enumerate(
            zip(self.computation_nodes, cross_computation_nodes)
        ):
            ziped_crosses = list(zip(cross_node_coords, cross_computation_nodes))
            same_nodes = [
                node == computation_node for node, computation_node in ziped_crosses
            ]
            bulk_nodes = [
                str(coord.idx) in map(str, computation_node.atoms(Indexed))
                for node, computation_node in ziped_crosses
            ]
            relevant_nodes = [
                (same_node or bulk_node)
                for same_node, bulk_node in zip(same_nodes, bulk_nodes)
            ]
            this_node = computation_node_coords[coord_idx]
            if all(relevant_nodes) and str(coord.idx) not in map(
                str, this_node.atoms()
            ):
                yield this_node

    def _node_to_domain(self, node_coords):
        domain = []
        for i, (coord, node_coord) in enumerate(zip(self.coordinates, node_coords)):
            if str(coord.idx) not in map(str, node_coord.atoms()):
                domain.append(Eq(coord.idx, node_coord))
            else:
                # get all the nodes on the same "cross"
                same_cross_nodes = list(self._get_filtered_cross(coord, node_coords))
                right_nodes = set(
                    [
                        node
                        for node in same_cross_nodes
                        if str(coord.N) in map(str, node.atoms())
                    ]
                )
                left_nodes = set(same_cross_nodes) - right_nodes

                left_node = max(left_nodes)
                right_node = (
                    min([node - coord.idx.upper for node in right_nodes])
                    + coord.idx.upper
                )

                domain.append((left_node < coord.idx) & (coord.idx < right_node))
        return reduce(and_, domain)

    @property
    def computation_domains(self):
        domains = []
        for coord_nodes in self.computation_nodes:
            domains.append(self._node_to_domain(coord_nodes))
        return domains

    def get_ghost_equation(self, ghost, node):
        unk = self.mapper[str(ghost.base)]
        indices = ghost.indices
        domains = self.get_domains(ghost)
        distances = self.get_distances(ghost)
        coords, domains, distances = zip(
            *[
                (coord, domain, distance)
                for coord, domain, distance in zip(unk.coords, domains, distances)
                if distance > 0
            ]
        )
        boundaries = [
            self.parsed_boundary_conditions[unk, coord]
            for coord, domain in zip(coords, domains)
        ]
        eqs, kinds = zip(
            *[
                (
                    bc.get(
                        domain,
                        domains={
                            coord: domain
                            for coord, domain in zip(
                                unk.coords, self.get_domains(ghost)
                            )
                        },
                        evaluation_node=ghost,
                        offset=-distance,
                    ),
                    getattr(bc, "kind_%s" % domain),
                )
                for bc, coord, domain, distance in zip(
                    boundaries, coords, domains, distances
                )
            ]
        )
        for eq, kind, domain, coord in zip(eqs, kinds, domains, coords):
            if kind != "periodic":
                eval_indices = [
                    index
                    if coord_idx != coord
                    else (coord.idx.lower if domain == "left" else coord.idx.upper)
                    for index, coord_idx in zip(indices, unk.coordinates)
                ]
                return eq(eval_indices)
            else:
                return eq(indices)

    def _extrapolate_coords(self, local_eq):
        outside_coords = [
            idx
            for idx in filter(self.is_outside, local_eq.atoms(Indexed))
            if isinstance(self.mapper[str(idx.base)], Coordinate)
        ]
        for indexed_coord in outside_coords:
            coord = self.mapper[str(indexed_coord.base)]
            index = indexed_coord.indices[0]
            distance = coord.distance_from_domain(index)
            domain = coord.domain(index)
            if domain == "left":
                rhs = coord.discrete[coord.idx.lower] - coord.step * distance
            else:
                rhs = coord.discrete[coord.idx.upper] + coord.step * distance
            logger.debug(
                "extrapolate outside coordinate, subs: {%s: %s}" % (indexed_coord, rhs)
            )
            local_eq = local_eq.subs(indexed_coord, rhs)
        return local_eq

    def _get_ghosts(self, local_eq):
        return [
            idx
            for idx in filter(self.is_outside, local_eq.atoms(Indexed))
            if isinstance(self.mapper[str(idx.base)], Unknown)
        ]

    # long running, should not be in property
    @cached_property
    def piecewise_system(self):
        piecewise_system = []
        for node_coord, domain in zip(self.computation_nodes, self.computation_domains):
            if node_coord in self.dirichlet_nodes:
                piecewise_system.append(Number(0))
                continue
            local_eq = self._extrapolate_coords(self(node_coord))
            ghosts = self._get_ghosts(local_eq)
            while ghosts:
                logger.debug("ghosts: %s" % ghosts)
                eqs = [self.get_ghost_equation(ghost, node_coord) for ghost in ghosts]
                solved = solve(eqs, ghosts, dict=True)
                local_eq = self._extrapolate_coords(local_eq.subs(solved[0]))
                ghosts = self._get_ghosts(local_eq)
            piecewise_system.append(local_eq)
        return tuple(piecewise_system)


def sign_distance(distance, domain):
    if domain == "bulk":
        return 0
    if domain == "left":
        return distance
    return -distance


def dict_product(d):
    keys = d.keys()
    for element in product(*d.values()):
        yield dict(zip(keys, element))


@attr.s
class PDESys:
    evolution_equations = attr.ib(
        type=typing.Sequence[str], converter=_convert_pde_list
    )
    unknowns = attr.ib(type=typing.Sequence[Unknown], converter=_convert_unk_list)
    parameters = attr.ib(
        type=typing.Sequence[Unknown], converter=_convert_unk_list, factory=list
    )
    coordinates = attr.ib(
        type=typing.Sequence[Coordinate],
        default=[],
        converter=_convert_coord_list,
        repr=False,
    )
    boundary_conditions = attr.ib(type=dict, factory=dict)
    subs = attr.ib(type=dict, factory=dict)
    # domains = attr.ib(type=dict, default=None, init=False, repr=False)

    def _coerce_equations(self):
        self.evolution_equations = [
            PDEquation(
                eq,
                self.unknowns,
                self.parameters,
                subs=self.subs,
                boundary_conditions=self.boundary_conditions,
            )
            for eq in self.evolution_equations.copy()
        ]
        self._apply_dirichlet()
        self.coordinates = sorted(
            set(chain(*[eq.coordinates for eq in self.evolution_equations.copy()]))
        )

    def _apply_dirichlet(self):
        for unk, pde in zip(self.unknowns, self.evolution_equations):
            for domains, coord_node in zip(pde.physical_domains, pde.node_coords):
                bcs = [
                    pde.parsed_boundary_conditions[(unk, coord)]
                    for coord in unk.coordinates
                ]
                kinds = [
                    getattr(bc, "kind_%s" % domain)
                    for bc, domain in zip(bcs, domains)
                    if domain != "bulk"
                ]
                if "Dirichlet" in kinds:
                    pde.dirichlet_nodes.append(coord_node)

    def _get_shapes(self):
        self.pivot = None

        gridinfo = dict(zip(self.coordinates, [coord.N for coord in self.coordinates]))

        shapes = []
        for unk in self.unknowns:
            gridshape = [
                (gridinfo[coord] if coord in unk.coordinates else 1)
                for coord in self.coordinates
            ]
            shapes.append(tuple(gridshape))
        sizes = [reduce(mul, shape) for shape in shapes]
        self.size = sum(sizes)

        self.shapes = dict(zip(self.unknowns, shapes))
        self.sizes = dict(zip(self.unknowns, sizes))

    def __attrs_post_init__(self):
        logger.info("processing pde system")
        logger.info("coerce equations...")
        self._t = Symbol("t")
        self._coerce_equations()
        self._get_shapes()
        logger.info("done")

    def _sort_indexed(self, indexed):
        unk_idx = [unk.name for unk in self.unknowns].index(str(indexed.args[0]))
        coords = self.unknowns[unk_idx].coordinates
        idxs = indexed.args[1:]
        idxs = [
            idxs[coords.index(coord)] if coord in coords else 0
            for coord in self.coordinates
        ]
        return [unk_idx, *idxs]

    def _filter_unk_indexed(self, indexed):
        return indexed.base in [unk.discrete for unk in self.unknowns]

    def _simplify_kron(self, *kron_args):
        kron = KroneckerDelta(*kron_args)
        return kron.subs({coord.N: oo for coord in self.coordinates})

    @cached_property
    def piecewise_system(self):
        return list(chain(*(pde.piecewise_system for pde in self)))

    @cached_property
    def jacobian_values(self):
        jacobian_values = []
        for expr in self.piecewise_system:
            wrts = list(filter(self._filter_unk_indexed, expr.atoms(Indexed)))
            diffs = [
                expr.diff(wrt).replace(KroneckerDelta, self._simplify_kron).n()
                for wrt in wrts
            ]
            jacobian_values.append(diffs)
        return jacobian_values

    @cached_property
    def jacobian_columns(self):
        jacobian_columns = []
        for expr in self.piecewise_system:
            wrts = list(filter(self._filter_unk_indexed, expr.atoms(Indexed)))
            grids = list(map(self._sort_indexed, wrts))
            jacobian_columns.append(grids)
        return jacobian_columns

    @property
    def unknowns_dict(self):
        return {unk.name: unk for unk in self.unknowns}

    @property
    def coordinates_dict(self):
        return {
            coord.name: coord
            for coord in set(chain(*[unk.coords for unk in self.unknowns]))
        }

    @property
    def parameters_dict(self):
        return {par.name: par for par in self.parameters}

    @property
    def mapper(self):
        return dict(
            **self.unknowns_dict, **self.parameters_dict, **self.coordinates_dict
        )

    @property
    def equation_dict(self):
        return {
            unk.name: equation
            for unk, equation in zip(self.unknowns, self.evolution_equations)
        }

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.evolution_equations[key]
        if isinstance(key, str):
            return self.equation_dict[key]
        raise KeyError(key)


def target_axis(derivative, wrt, axis):
    return wrt == axis.symbol


@attr.s
class BoundaryCondition:
    unknown = attr.ib(type=Unknown, converter=Unknown)
    axis = attr.ib(type=Coordinate, converter=Coordinate)
    pde = attr.ib(PDEquation)
    bcs = attr.ib(type=typing.Optional[typing.Union[str, typing.Tuple[str, str]]])
    left = attr.ib(type=PDEquation, init=False, repr=False)
    right = attr.ib(type=PDEquation, init=False, repr=False)
    kind_left = attr.ib(type=str, init=False)
    kind_right = attr.ib(type=typing.Union[typing.Tuple[str, str], str], init=False)

    @bcs.default
    def noflux_default(self):
        return ("noflux", "noflux")

    @bcs.validator
    def bcs_validator(self, attribute, value):
        if value in ["periodic", "noflux"]:
            return True
        if not isinstance(value, str) and len(value) == 2:
            return True

    def _build_bcs(self, bcs):
        if bcs == "periodic":
            left = self._build_periodic("left")
            right = self._build_periodic("right")
            return (left, "periodic"), (right, "periodic")
        if bcs == "noflux":
            bcs = ("noflux", "noflux")
        return (self._build_bc(bc) for bc in bcs)

    def _build_bc(self, eq):
        kind = None
        if eq is None or eq == "noflux":
            eq = "d%s%s" % (self.axis.name, self.unknown.name)
        if eq == "dirichlet":
            eq = "d%s%s" % (self.axis.name, self.unknown.name)
            kind = "Dirichlet"

        return eq, kind or "Ghost"

    def _build_periodic(self, side):
        unk = self.unknown
        axis = self.axis
        if side == "left":
            substitute = axis.idx + axis.N
        else:
            substitute = axis.idx - axis.N
        return unk.discrete_i - unk.discrete_i.subs(axis.idx, substitute)

    def __attrs_post_init__(self):
        self.bcs = self.bcs or "noflux"
        (self.left, self.kind_left), (self.right, self.kind_right) = self._build_bcs(
            self.bcs
        )

    def evaluate_side_patterns(self, node, coord, domain):
        if coord == self.axis:
            return
        if str(node) == str(coord.idx):
            return
        if domain == "bulk" and str(coord.idx) in map(str, node.atoms(Symbol)):
            this_offset = node - coord.idx
            if this_offset > 0:
                scheme = "left"
            else:
                scheme = "right"
        elif domain == "bulk":
            return
        else:
            if domain == "right":
                scheme = "left"
            else:
                scheme = "right"
        return FiniteDifferenceScheme(
            scheme=scheme, pattern=partial(target_axis, axis=coord)
        )

    def get(self, side, domains, evaluation_node, offset=None):
        if getattr(self, "kind_%s" % side) == "periodic":
            return PDEquation(self._build_periodic(side), self.unknown, raw=True)
        if offset is None:
            offset = 1

        if side == "left":
            scheme = "right"
            main_offset = offset
        else:
            scheme = "left"
            main_offset = -offset

        main_scheme = FiniteDifferenceScheme(
            scheme=scheme,
            offset=main_offset,
            pattern=partial(target_axis, axis=self.axis),
        )

        schemes = [main_scheme]

        for node, (coord, domain) in zip(evaluation_node.indices, domains.items()):
            scheme = self.evaluate_side_patterns(node, coord, domain)
            if scheme:
                schemes.append(scheme)

        return PDEquation(
            getattr(self, side),
            self.pde.unknowns,
            self.pde.parameters,
            schemes=schemes[::-1],
        )



#!/usr/bin/env python
# coding=utf-8

import sys
import warnings
from copy import deepcopy
from itertools import chain

import forge
import numpy as np
import cloudpickle

# from loguru import logger
from xarray import Dataset

from .core.backends import get_backend
from .core.grid_builder import GridBuilder
from .core.system import PDESys
from .simulation import Simulation, null_hook


sys.setrecursionlimit(40000)
EPS = 1e-6


def create_fields(coords, unks, pars):
    unk_names = [unk.name for unk in unks]
    ivar_names = [coord.name for coord in coords]
    par_names = [par.name for par in pars]

    @forge.sign(
        *[forge.kwarg(name=name) for name in chain(ivar_names, unk_names, par_names)]
    )
    def Fields(**kwargs):
        """Create a xarray Dataset as expected by the model input."""
        return Dataset(
            data_vars={
                unk.name: (
                    unk.c_names,
                    np.broadcast_to(
                        kwargs[unk.name], (kwargs[coord].size for coord in unk.c_names)
                    ).copy(),
                )
                for unk in chain(unks, pars)
            },
            coords={coord.name: kwargs[coord.name] for coord in coords},
        )

    return Fields


class Model:
    r"""Contain finite difference approximation and routine of the dynamical system.

    Take a mathematical form as input, use Sympy to transform it as a symbolic
    expression, perform the finite difference approximation and expose theano
    optimized routine for both the right hand side of the dynamical system and
    Jacobian matrix approximation.

    Parameters
    ----------
    evolution_equations : Union[Sequence[str], str]
        the right hand sides of the partial differential equations written
        as :math:`\\frac{\\partial U}{\\partial t} = F(U)`, where the spatial
        derivative can be written as `dxxU` or `dx(U, 2)` or with the sympy
        notation `Derivative(U, x, x)`
    unknowns : Union[Sequence[str], str]
        the dependent variables with the same order as the temporal
        derivative of the previous arg.
    parameters : Optional[Union[Sequence[str], str]]
        list of the parameters. Can be feed with a scalar of an array with
        the same size. They can be derivated in space as well.
    boundary_conditions : Optional[Union[str, Dict[str, Dict[str, Tuple[str, str]]]]]
        Can be either "noflux" (default behavior), "periodic", or a dictionary that
        follow this structure :
        {dependent_var: {indep_var: (left_boundary, right_boundary)}}, the boundaries
        being written as residual (rhs - lhs = 0).
        For example, an hybrid Dirichlet / Neumann flux will be written as
        {"U": {"x": (dxU - 2, U - 3)}}. If a boundary is None, a nul flux will be
        applied.
    subs : Optional[Dict[str, str]]
        Substitution dictionnary, useful to write complex systems. The key of this
        dictionnary will be substitued everywhere in the evolution equations as well
        as in boundary conditions.
    backend: str, default to "numpy"
        A registered backend that will take the discretized system and expose the evolution
        equations routine ``F`` as well as an optional Jacobian routine ``J``. If the later is
        not provided by the backend, the implicit methods will not be available.
        TODO: make an efficient jacobian approximation for that case, given the bandwidth.
        For now, there is the ``numpy`` backend, that provide efficient (but not optimal)
        computation with minimal dependencies, and ``numba`` backend that provide a LLVM based
        routine that allow faster parallel computation at the costs of a (sometime long) warm-up.
        The numba one is less tested and stable, but provide a potential huge speed-up for explicit
        methods.
    backend_kwargs: Optional[Dict] : supplementary arguments provided to the backend.

    Examples
    --------
    A simple diffusion equation:

    >>> from skfdiff import Model
    >>> model = Model("k * dxxU", "U", "k")

    A coupled system of convection-diffusion equation:

    >>> from skfdiff import Model
    >>> model = Model(["k1 * dxxU - c1 * dxV",
    ...                "k2 * dxxV - c2 * dxU",],
    ...                ["U", "V"], ["k1", "k2", "c1", "c2"])

    A 2D pure advection model, without / with upwind scheme.

    >>> from skfdiff import Model
    >>> model = Model("c_x * dxU + c_y * dyU", "U(x, y)",
    ...               ["c_x", "c_y"])
    >>> model = Model("upwind(c_x, U, x, 2) + upwind(c_y, U, y, 2)",
    ...               "U(x, y)",
    ...               ["c_x", "c_y"])


    A 2D diffusion model, with hybrid boundary conditions (Dirichlet, Neuman, Robin and No-flux).

    >>> from skfdiff import Model
    >>> model = Model("dxxU + dyyU", "U(x, y)",
    ...               boundary_conditions={("U", "x"): ("U - 3", "dxU + 5"),
    ...                                    ("U", "y"): ("dyU - (U - 3)", None)
    ...                                   })

    """  # noqa

    def __init__(
        self,
        evolution_equations,
        unknowns,
        parameters=None,
        coordinates=None,
        boundary_conditions=None,
        subs=None,
        backend="numpy",
        backend_kwargs=None,
    ):
        if parameters is None:
            parameters = []
        if coordinates is None:
            coordinates = []
        if boundary_conditions is None:
            boundary_conditions = {}
        if subs is None:
            subs = {}
        if backend_kwargs is None:
            backend_kwargs = {}

        self.evolution_equations = evolution_equations
        self.unknowns = deepcopy(unknowns)
        self.parameters = deepcopy(parameters)
        self.coordinates = deepcopy(coordinates)
        self.boundary_conditions = deepcopy(boundary_conditions)
        self.subs = deepcopy(subs)

        self.pdesys = PDESys(
            evolution_equations=self.evolution_equations,
            unknowns=self.unknowns,
            coordinates=self.coordinates,
            parameters=self.parameters,
            boundary_conditions=self.boundary_conditions,
            subs=self.subs,
        )

        self.grid_builder = GridBuilder(self.pdesys)

        self.backend = get_backend(backend)(
            self.pdesys, self.grid_builder, **backend_kwargs
        )

    @property
    def Fields(self):
        """Base for structured fields stored in a xarray."""
        return create_fields(
            self.pdesys.coordinates, self.pdesys.unknowns, self.pdesys.parameters
        )

    def F(self, fields, t=0):
        """Compute the right hand side of the dynamical system"""
        return self.backend.F(fields, t=t)

    def J(self, fields, t=0):
        """Compute the Jacobian of the dynamical system as sparce csc matrix"""
        try:
            return self.backend.J(fields, t=t)
        except AttributeError:
            raise AttributeError(
                "current backend %s lack of routine to compute the jacobian matrix. "
                "Only explicite scheme will be working." % self.backend.name
            )

    def U_from_fields(self, fields, t=0):
        """Get a fields and return the data as a single flat vector."""
        return self.grid_builder.U_from_fields(fields, t=t)

    def fields_from_U(self, U, fields, t=0):
        """take a solution ad single flat vector and return structured fields."""
        return self.grid_builder.fields_from_U(U, fields, t=t)

    def __getstate__(self):
        return {key: cloudpickle.dumps(value) for key, value in self.__dict__.items()}

    def __setstate__(self, state):
        self.__dict__.update(
            {key: cloudpickle.loads(value) for key, value in state.items()}
        )

    def __repr__(self):
        if not self.boundary_conditions:
            bdc = "noflux"
        elif isinstance(self.boundary_conditions, dict):
            bdc = "\n".join(self.boundary_conditions.items())
        else:
            bdc = self.boundary_conditions
        return """System
-------
{pdesys}

Boundary condition
------------------
{bdc}

backend
-------
{backend}
""".format(
            pdesys="\n\n".join(
                [
                    "dt%s = %s" % (unk, pde.equation)
                    for unk, pde in zip(self.unknowns, self.pdesys)
                ]
            ),
            bdc=bdc,
            backend=self.backend.name,
        )

    def fields_template(self, *args, **kwargs):
        warnings.warn(
            "This interface is deprecied, use model.Fields instead. Will be removed in the 0.7.0",
            DeprecationWarning,
        )
        return self.Fields(*args, **kwargs)

    def init_simulation(
        self,
        fields,
        dt,
        t=0,
        tmax=None,
        id=None,
        hook=null_hook,
        scheme="RODASPR",
        time_stepping=True,
        **kwargs
    ):
        """Init a skfdiff Simulation. See skfdiff.simulation.Simulation docstring
        for further details.
        """
        return Simulation(
            model=self,
            fields=fields,
            dt=dt,
            t=t,
            tmax=tmax,
            id=id,
            hook=hook,
            scheme=scheme,
            time_stepping=time_stepping,
            **kwargs
        )



#!/usr/bin/env python
# coding=utf-8

from operator import attrgetter


import attr
import numba
import numpy
from jinja2 import Template
from loguru import logger
from scipy.sparse import csc_matrix
from sympy import Indexed, KroneckerDelta, Symbol, lambdify, oo
from sympy.printing.lambdarepr import NumPyPrinter
import numpy as np
from cachetools import cachedmethod

from .base_backend import Backend, register_backend
from .numpy_backend import np_Heaviside, np_Max, np_Min


def np_depvar_printer(unk):
    return Symbol(str(unk.discrete))


def np_ivar_printer(coord):
    return (Symbol(str(coord.discrete)), Symbol(str(coord.idx)), coord.step, coord.N)


# @numba.jit(nopython=True, parallel=True, fastmath=True)
@numba.vectorize(
    [numba.float64(numba.float64, numba.float64)], nopython=True, fastmath=True
)
def Min(a, b):
    if a < b:
        return a
    else:
        return b


# @numba.jit(nopython=True, parallel=True, fastmath=True)
@numba.vectorize([numba.float64(numba.float64, numba.float64)], nopython=True)
def Max(a, b):
    if a > b:
        return a
    else:
        return b


@numba.vectorize([numba.float64(numba.float64)], nopython=True)
def Heaviside(x):
    if x < 0:
        return 0
    else:
        return 1


class EnhancedNumPyPrinter(NumPyPrinter):
    def _print_Idx(self, idx, **kwargs):
        return str(idx)

    def _print_Indexed(self, indexed, **kwargs):
        return str(indexed)

    def _print_Min(self, expr):
        return "{0}({1})".format(
            self._module_format("Min"), ",".join(self._print(i) for i in expr.args)
        )

    def _print_Max(self, expr):
        return "{0}({1})".format(
            self._module_format("Max"), ",".join(self._print(i) for i in expr.args)
        )


@register_backend
@attr.s()
class NumbaBackend(Backend):
    name = "numba"
    parallel = attr.ib(default=False, type=bool)
    fastmath = attr.ib(default=True, type=bool)
    nogil = attr.ib(default=True, type=bool)
    printer = EnhancedNumPyPrinter()

    def _convert_inputs(self):
        self.ndim = len(self.system.coordinates)
        self.unks = list(map(np_depvar_printer, self.system.unknowns))
        self.pars = list(map(np_depvar_printer, self.system.parameters))
        self.coords, self.idxs, self.steps, self.sizes = zip(
            *map(np_ivar_printer, self.system.coordinates)
        )
        self.idxs_map = [
            tuple([self.system.coordinates.index(coord) for coord in unk.coordinates])
            for unk in self.system.unknowns
        ]

        self.shapes = [self.system.shapes[unk] for unk in self.system.unknowns]

        self.inputs = [
            self.system._t,
            *self.unks,
            *self.pars,
            *self.coords,
            *self.steps,
            *self.sizes,
        ]
        self.inputs_cond = [*self.idxs, *self.sizes]

    def _build_evolution_equations(self):
        template = Template(
            """
@numba.jit(nopython=True, parallel={{parallel}}, fastmath={{fastmath}}, nogil={{nogil}})
def compute_F_vector({% for item in var_names %}{{ item }}, {% endfor %}grid):
    N = grid.shape[0]
    F = numpy.empty(N)
    for i in numba.prange(N):
        didx, {% for item in var_unpacking %}{{ item }}, {% endfor %}domain, flatidx = grid[i]
        {% for i, expr in exprs %}
        if domain == {{ i }}:
            F[flatidx] = {{ expr }}{% endfor %}
    return F
"""  # noqa
        )
        var_names = self.inputs
        var_unpacking = self.idxs
        self._F_routine_str = template.render(
            var_names=var_names,
            var_unpacking=var_unpacking,
            exprs=enumerate(
                [self.printer.doprint(expr) for expr in self.system.piecewise_system]
            ),
            parallel=self.parallel,
            fastmath=self.fastmath,
            nogil=self.nogil,
        )

        # the routine will be added to the class scope
        exec(self._F_routine_str, globals(), self.__dict__)

    def F(self, fields, t=0):
        unks = [
            fields[varname].values
            for varname in [unk.name for unk in self.system.unknowns]
        ]
        pars = [
            fields[varname].values
            for varname in [par.name for par in self.system.parameters]
        ]
        coords = [
            fields[varname].values
            for varname in [coord.name for coord in self.system.coordinates]
        ]
        sizes = [coord.size for coord in coords]
        grid = self.grid_builder.compute_gridinfo(sizes)
        steps = self.grid_builder.compute_steps(
            tuple(sizes), tuple([coord.ptp() for coord in coords])
        )

        return self.compute_F_vector(t, *unks, *pars, *coords, *steps, *sizes, grid)

    def _build_jacobian(self):

        template = Template(
            """
@numba.jit(nopython=True, parallel={{parallel}}, fastmath={{fastmath}}, nogil={{nogil}})
def compute_jacobian_values({% for item in var_names %}{{ item }},
                            {% endfor %}subgrids, data_size):
    N = len(subgrids)
    data = numpy.zeros(data_size)
    cursor = 0
    jacs_lenght = {{ jacs_len }}

    cursors = [0]
    for i in range(len(subgrids)):
        grid = subgrids[i]
        jl = jacs_lenght[i]
        cursors.append(grid.shape[0] * jl + cursors[-1])
    for i in numba.prange(N):
        cursor = cursors[i]
        grid = subgrids[i]
        M = grid.shape[0]
        {% for jacs in full_jacs %}
        {{ "if" if loop.index==1 else "elif" }} i == {{loop.index0}}: {% if jacs %}{% for jac_func in jacs %}
            next_cursor = cursor + M
            for j in numba.prange(M): {% for item in var_unpacking %}
                {{ item }} = grid[j, {{ loop.index }}]{% endfor %}
                data[cursor + j] = {{ jac_func }}
            cursor = next_cursor {% endfor %}{% else %}
            pass
        {% endif %} {% endfor %}

    return data
        """
        )

        var_names = self.inputs
        var_unpacking = self.idxs
        self._J_routine_str = template.render(
            var_names=var_names,
            var_unpacking=var_unpacking,
            full_jacs=[
                [self.printer.doprint(jac) for jac in jacs]
                for jacs in self.system.jacobian_values
            ],
            jacs_len=[len(jacs) for jacs in self.system.jacobian_values],
            parallel=self.parallel,
            fastmath=self.fastmath,
            nogil=self.nogil,
        )
        logger.debug("generated routine: ")

        # the routine will be added to the class scope
        exec(self._J_routine_str, globals(), self.__dict__)

        # the routine is the same as the numpy backend: they only have to be executed
        # for the first evaluation: no need to rewrite a numba (complex) implementation.

    @cachedmethod(cache=attrgetter("_cache_jac_coord"))
    def compute_jacobian_coordinates(self, *sizes):
        self.jacobian_columns = [
            [lambdify(self.inputs_cond, col) for col in columns]
            for columns in self.system.jacobian_columns
        ]
        subgrids = self.grid_builder.compute_subgrids(sizes)
        system_sizes = self.grid_builder.compute_sizes(sizes)
        system_size = sum(system_sizes)

        rows_list = []
        cols_list = []
        for grid, jac_cols in zip(subgrids, self.jacobian_columns):
            for col_func in jac_cols:
                cols_ = np.zeros((grid.shape[0], self.ndim + 1), dtype="int32")
                cols = col_func(*grid[:, 1:-2].T, *sizes)
                cols = np.stack(
                    [np.broadcast_to(col, cols_.shape[:-1]) for col in cols]
                )

                flat_cols = self.grid_builder.compute_flat_from_idxs(cols.T, sizes)

                rows_list.extend(grid[:, -1].reshape((-1,)))
                cols_list.extend(flat_cols.reshape((-1,)))
        rows = np.array(rows_list)
        cols = np.array(cols_list)

        perm = np.argsort(cols)
        perm_rows = rows[perm]
        perm_cols = cols[perm]
        count = np.zeros((system_size + 1), dtype="int32")
        uq, cnt = np.unique(perm_cols, False, False, True)
        count[uq + 1] = cnt
        indptr = np.cumsum(count)
        return rows, cols, perm_rows, indptr, perm, (system_size, system_size)

    def J(self, fields, t=0):
        unks = [
            fields[varname].values
            for varname in [unk.name for unk in self.system.unknowns]
        ]
        pars = [
            fields[varname].values
            for varname in [par.name for par in self.system.parameters]
        ]
        coords = [
            fields[varname].values
            for varname in [coord.name for coord in self.system.coordinates]
        ]
        sizes = [coord.size for coord in coords]

        steps = self.grid_builder.compute_steps(
            tuple(sizes), tuple([coord.ptp() for coord in coords])
        )
        subgrids = self.grid_builder.compute_subgrids(sizes)
        data_size = sum(
            [
                subgrid.shape[0] * len(jacs)
                for subgrid, jacs in zip(subgrids, self.system.jacobian_columns)
            ]
        )
        data = self.compute_jacobian_values(
            t, *unks, *pars, *coords, *steps, *sizes, subgrids, data_size
        )
        _, _, perm_rows, indptr, perm, shape = self.compute_jacobian_coordinates(*sizes)

        return csc_matrix((data[perm], perm_rows, indptr), shape=shape)

    def __attrs_post_init__(self):
        self._cache_jac_coord = {}
        logger.info("numba backend: convert_inputs...")
        self._convert_inputs()
        logger.info("numba backend: build_evolution_equations...")
        self._build_evolution_equations()
        logger.info("numba backend: build_jacobian...")
        self._build_jacobian()



#!/usr/bin/env python
# coding=utf-8

from functools import partial, wraps
from itertools import chain
from warnings import warn

import holoviews as hv
from boltons.iterutils import remap
from loguru import logger
from sympy import Dummy, Idx, Indexed, solve
from tqdm import tqdm, tqdm_notebook

tqdm = tqdm


def generate_dummy_map(exprs):
    to_dummify = set(chain(*[expr.atoms(Indexed, Idx) for expr in exprs]))
    dummy_map = {var: Dummy() for var in to_dummify}
    reverse_dummy_map = {dummy: var for var, dummy in dummy_map.items()}
    return dummy_map, reverse_dummy_map


@wraps(solve)
def solve_with_dummy(exprs, vars_, *args, **kwargs):
    def visit_dummy(subs, path, key, value):
        try:
            value = value.subs(subs)
        except AttributeError:
            pass
        try:
            key = key.subs(subs)
        except AttributeError:
            pass
        return key, value

    dummy_map, reverse_dummy_map = generate_dummy_map(exprs)
    dummy_exprs = [expr.subs(dummy_map) for expr in exprs]
    dummy_vars = [var.subs(dummy_map) for var in vars_]
    dummy_sol = solve(dummy_exprs, dummy_vars)
    sol = remap(dummy_sol, visit=partial(visit_dummy, reverse_dummy_map))
    return sol


def _enable_tqdm():
    global tqdm
    try:
        import ipywidgets  # noqa

        tqdm = tqdm_notebook
    except ImportError:
        warn("ipywidgets not installed, skipping tqdm_notebook")


def _enable_bokeh(backends):
    try:
        import bokeh  # noqa

        backends.append("bokeh")
    except ImportError:
        warn("bokeh not installed, skipping this backend")
    return backends


def _enable_matplotlib(backends):
    try:
        import matplotlib  # noqa

        backends.append("matplotlib")
    except ImportError:
        warn("bokeh not installed, skipping this backend")
    return backends


def enable_notebook(enable_tqdm=True, enable_bokeh=True, enable_matplotlib=True):
    if enable_tqdm:
        _enable_tqdm()
    backends = []
    if enable_bokeh:
        backends = _enable_bokeh(backends)
    if enable_matplotlib:
        backends = _enable_matplotlib(backends)

    if backends:
        import logging  # noqa

        hv.notebook_extension(*backends)
        # this function has an annoying side effect and somehow
        # use the root log, add a StreamHandler and put the log
        # level to INFO. This will reset that undesired behaviour.
        root_logging = logging.getLogger()
        root_logging.handlers = []
        root_logging.setLevel("WARNING")


def transform_into_hashable(key):
    try:
        hash(key)
        return key
    except TypeError:
        if isinstance(key, (list, set)):
            return tuple(sorted(key))
        if isinstance(key, dict):
            return tuple(sorted(key.items()))
        return str(key)



#!/usr/bin/env python
# coding=utf-8
# flake8: noqa

from loguru import logger

from .utils import enable_notebook  # noqa
from .core import temporal_schemes as schemes  # noqa
from .core.temporal_schemes import Stationnary as StationnarySolver
from .model import Model  # noqa
from .simulation import Simulation  # noqa
from .core.system import PDESys, PDEquation  # noqa
from .core.grid_builder import GridBuilder  # noqa
from .core import backends

from .plugins.container import Container  # noqa
from .plugins.displays import Display  # noqa

retrieve_container = Container.retrieve
display_fields = Display.display_fields

from ._version import get_versions  # noqa

__version__ = get_versions()["version"]
del get_versions

import sys

logger.disable(__name__)



#!/usr/bin/env python
# coding=utf-8

from functools import partial
from itertools import accumulate
from typing import Tuple

import numpy as np
from cachetools import Cache, cachedmethod
from cachetools.keys import hashkey
from loguru import logger
from sympy import Matrix, lambdify
import cloudpickle

from .system import PDESys


def retrieve_cache(key, self):
    try:
        return self._cache[key]
    except KeyError:
        self._cache[key] = Cache(maxsize=128)
        return self._cache[key]


def list_to_tuple_key(*args, **kwargs):
    args = [tuple(arg) if isinstance(arg, list) else arg for arg in args]
    kwargs = {
        key: tuple(value) if isinstance(value, list) else value
        for key, value in kwargs.items()
    }
    key = hashkey(*args, **kwargs)
    return key


class GridBuilder:
    def __init__(self, system: PDESys):
        self.system = system
        self._cache = {}
        self._convert_inputs()
        self._lambda_shapes = [
            lambdify(self.sizes, shape, modules="numpy") for shape in self.shapes
        ]

        # lambdify the size computation for each dep. var. (shape product)
        self._lambda_sizes = [
            lambdify(self.sizes, self.system.sizes[unk], modules="numpy")
            for unk in self.system.unknowns
        ]

        self._lambda_conds = []

        for pde in self.system:
            conds = pde.computation_domains
            lambda_cond = lambdify(self.inputs_cond, Matrix(conds), modules=["numpy"])
            self._lambda_conds.append(lambda_cond)

    def __getstate__(self):
        return {key: cloudpickle.dumps(value) for key, value in self.__dict__.items()}

    def __setstate__(self, state):
        self.__dict__.update(
            {key: cloudpickle.loads(value) for key, value in state.items()}
        )

    def _convert_inputs(self):
        self.ndim = len(self.system.coordinates)
        self.unks = [unk.symbol for unk in self.system.unknowns]
        self.sizes = [coord.N for coord in self.system.coordinates]
        self.idxs = [coord.idx for coord in self.system.coordinates]
        self.idxs_map = [
            tuple([self.system.coordinates.index(coord) for coord in unk.coordinates])
            for unk in self.system.unknowns
        ]

        self.shapes = [self.system.shapes[unk] for unk in self.system.unknowns]

        self.inputs_cond = [*self.idxs, *self.sizes]

    @cachedmethod(
        cache=partial(retrieve_cache, "build_flat_from_idxs_mapper"),
        key=list_to_tuple_key,
    )
    def build_flat_from_idxs_mapper(self, sizes: Tuple[int]):
        mapper = {}
        gridinfo = self.compute_gridinfo(sizes)
        for row in gridinfo:
            mapper[tuple(row[:-2])] = row[-1]
        return mapper

    @cachedmethod(
        cache=partial(retrieve_cache, "build_idxs_from_flat_mapper"),
        key=list_to_tuple_key,
    )
    def build_idxs_from_flat_mapper(self, sizes: Tuple[int]):
        mapper = {}
        gridinfo = self.compute_gridinfo(sizes)
        for row in gridinfo:
            mapper[int(row[-1])] = row[:-2]
        return mapper

    def compute_flat_from_idxs(self, idxs, sizes: Tuple[int]):
        mapper = self.build_flat_from_idxs_mapper(sizes)
        if np.array(idxs).ndim == 1:
            return mapper[tuple(idxs)]
        else:
            return np.array([mapper[tuple(idx)] for idx in idxs])

    def compute_idxs_from_flat(self, flatindex, sizes: Tuple[int]):
        mapper = self.build_idxs_from_flat_mapper(sizes)
        if np.array(flatindex).ndim == 0:
            return mapper[int(flatindex)]
        else:
            return np.array([mapper[int(idx)] for idx in flatindex])

    def compute_pivot_idx(self, shapes: Tuple[int]):
        if np.alltrue(shapes == shapes[0]):
            return np.arange(self.ndim)
        return np.argsort(np.sum(shapes, axis=0))[::-1]

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_shapes"), key=list_to_tuple_key
    )
    def compute_shapes(self, sizes: Tuple[int]):
        """
        compute the shapes of all the model fields.
        """
        return tuple([lambda_shape(*sizes) for lambda_shape in self._lambda_shapes])

    @cachedmethod(cache=partial(retrieve_cache, "compute_steps"), key=list_to_tuple_key)
    def compute_steps(self, sizes, lenghts):
        """
        compute the stepsize of all the independent variables.
        """
        return tuple([lenght / (size - 1) for lenght, size in zip(lenghts, sizes)])

    @cachedmethod(cache=partial(retrieve_cache, "compute_sizes"), key=list_to_tuple_key)
    def compute_sizes(self, sizes: Tuple[int]):
        """
        compute the sizes of all the model fields.
        """
        return tuple([lambda_size(*sizes) for lambda_size in self._lambda_sizes])

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_indices"), key=list_to_tuple_key
    )
    def compute_indices(self, sizes: Tuple[int]):
        """
        compute grid indices of all the model fields.
        """
        shapes = self.compute_shapes(sizes)
        return [np.indices(np.stack(shape)) for shape in shapes]

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_domains"), key=list_to_tuple_key
    )
    def compute_domains(self, sizes: Tuple[int]):
        """
        compute grid indices of all the model fields.
        """
        indices = self.compute_indices(sizes)
        cursor = 0
        domains = []
        for lambda_cond, indice in zip(self._lambda_conds, indices):
            cond_grid = lambda_cond(*indice, *sizes)
            domain_grid = np.select(
                cond_grid, np.arange(cursor, cursor + cond_grid.shape[0])
            ).squeeze()
            cursor = domain_grid.max() + 1
            domains.append(domain_grid)
        return domains

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_gridinfo"), key=list_to_tuple_key
    )
    def compute_gridinfo(self, sizes: Tuple[int]):
        system_sizes = self.compute_sizes(sizes)
        system_size = sum(system_sizes)
        shapes = self.compute_shapes(sizes)
        indices = self.compute_indices(sizes)
        domains = self.compute_domains(sizes)
        ndim = self.ndim

        pivot_idx = self.compute_pivot_idx(shapes)
        cursors = list(accumulate([0, *system_sizes]))

        unk_info = np.zeros((system_size,), dtype="int32")
        idxs_info = np.zeros((ndim, system_size), dtype="int32")
        domains_info = np.zeros((system_size,), dtype="int32")

        for i, (cursor, size, indice, domain) in enumerate(
            zip(cursors, system_sizes, indices, domains)
        ):
            unk_info[cursor : cursor + size] = np.repeat(i, size)
            idxs_info[:, cursor : cursor + size] = indice.reshape((-1, size))
            domains_info[cursor : cursor + size] = domain.flatten()

        full_grid = np.concatenate([idxs_info[pivot_idx, :], unk_info[None, :]], axis=0)

        perm_vector = np.lexsort(full_grid[::-1])

        permuted_idxs_info = idxs_info[:, perm_vector]
        permuted_unk_info = unk_info[None, perm_vector]
        permuted_domains_info = domains_info[None, perm_vector]
        flatten_idx = np.arange(system_size, dtype="int32")

        gridinfo = np.concatenate(
            [
                permuted_unk_info,
                permuted_idxs_info,
                permuted_domains_info,
                flatten_idx[None, :],
            ],
            axis=0,
        ).T

        return gridinfo

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_flat_maps"), key=list_to_tuple_key
    )
    def compute_flat_maps(self, sizes: Tuple[int]):
        shapes = self.compute_shapes(sizes)
        gridinfo = self.compute_gridinfo(sizes)
        flatmaps = [
            np.compress(gridinfo[:, 0] == i, gridinfo[:, -1], axis=0).reshape(shape)
            for i, shape in enumerate(shapes)
        ]
        return flatmaps

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_subgrids"), key=list_to_tuple_key
    )
    def compute_subgrids(self, sizes: Tuple[int]):
        gridinfo = self.compute_gridinfo(sizes)
        domains = sorted(
            set(
                np.concatenate(
                    [domain.flatten() for domain in self.compute_domains(sizes)]
                ).tolist()
            )
        )
        condlists = [gridinfo[:, -2] == i for i in domains]
        subgrids = [np.compress(condlist, gridinfo, axis=0) for condlist in condlists]
        return subgrids

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_flat_to_unks"), key=list_to_tuple_key
    )
    def compute_flat_to_unks(self, sizes: Tuple[int]):
        gridinfo = self.compute_gridinfo(sizes)
        idxs_grids = [
            self.compute_flat_from_idxs(
                np.compress(gridinfo[:, 0] == i, gridinfo[:, :-2], axis=0), sizes
            )
            for i in range(len(self.unks))
        ]

        return idxs_grids

    @cachedmethod(
        cache=partial(retrieve_cache, "compute_unks_to_flat"), key=list_to_tuple_key
    )
    def compute_unks_to_flat(self, sizes: Tuple[int]):
        system_sizes = self.compute_sizes(sizes)
        system_size = sum(system_sizes)
        ptrs = self.compute_idxs_from_flat(np.arange(system_size, dtype="int32"), sizes)
        return ptrs

    @cachedmethod(cache=partial(retrieve_cache, "compute_idxs"), key=list_to_tuple_key)
    def compute_idxs(self, sizes: Tuple[int]):
        system_sizes = self.compute_sizes(sizes)
        indices = self.compute_indices(sizes)
        indices = [
            indice.reshape(-1, size) for indice, size in zip(indices, system_sizes)
        ]
        unk_idx = np.concatenate(
            [np.repeat(i, size) for i, size in enumerate(system_sizes)]
        )
        idxs = np.concatenate(
            [indices[i] for i, size in enumerate(system_sizes)], axis=1
        )
        return np.stack([unk_idx, *idxs], axis=0)

    @cachedmethod(
        cache=partial(retrieve_cache, "get_relevant_ptrs"), key=list_to_tuple_key
    )
    def get_relevant_ptrs(self, i: int, sizes: Tuple[int]):
        system_sizes = self.compute_sizes(sizes)
        system_size = sum(system_sizes)
        shapeinfos = self.compute_shapes(sizes)
        pivot_idx = self.compute_pivot_idx(shapeinfos)
        idxs = np.arange(system_size)
        ptrs = self.compute_unks_to_flat(sizes)
        relevant_idxs = np.extract(ptrs[:, 0] == i, idxs)
        relevant_ptrs = ptrs[relevant_idxs, 1:].T[pivot_idx]
        return relevant_idxs, tuple(relevant_ptrs)

    def U_routine(self, unks, sizes: Tuple[int]):
        system_sizes = self.compute_sizes(sizes)
        system_size = sum(system_sizes)
        U = np.empty(system_size)

        for i, unk in enumerate(unks):
            relevant_idxs, relevant_ptrs = self.get_relevant_ptrs(i, sizes)
            U[relevant_idxs] = unk[relevant_ptrs]
        return U

    def U_from_fields(self, fields, t=0):
        unk_names = [unk.name for unk in self.system.unknowns]
        ivar_names = [coord.name for coord in self.system.coordinates]
        unks = [fields[varname] for varname in unk_names]
        sizes = [fields[varname].size for varname in ivar_names]
        shapeinfos = self.compute_shapes(sizes)
        pivot_idx = self.compute_pivot_idx(shapeinfos)
        coords = list(map(str, np.array(self.system.coordinates)[pivot_idx]))
        unks = [
            unk.expand_dims(list(set(ivar_names).difference(set(unk.dims))))
            .transpose(*coords)
            .values
            for unk in unks
        ]

        return self.U_routine(unks, sizes)

    def fields_routine(self, U, sizes: Tuple[int]):
        shapes = self.compute_shapes(sizes)
        idxs_grids = self.compute_flat_to_unks(sizes)
        pivots = self.compute_pivot_idx(shapes)

        return [
            U[grid].reshape(np.array(shape)[pivots])
            for grid, shape in zip(idxs_grids, shapes)
        ]

    def fields_from_U(self, U, fields, t=None):
        varnames = [unk.name for unk in self.system.unknowns]
        fields = fields.copy()

        coords = [
            fields[varname]
            for varname in [coord.name for coord in self.system.coordinates]
        ]
        sizes = [coord.size for coord in coords]
        shapes = self.compute_shapes(sizes)
        pivots = self.compute_pivot_idx(shapes)
        unks = self.fields_routine(U, sizes)
        sys_ivars = self.system.coordinates
        for varname, unk, coords in zip(
            varnames, unks, [unk.coordinates for unk in self.system.unknowns]
        ):
            pivot_ordered_coords = [
                sys_ivars[i].name for i in pivots if sys_ivars[i] in coords
            ]
            fields[varname] = pivot_ordered_coords, unk.squeeze()
            fields[varname] = fields[varname].transpose(
                *[coord.name for coord in sys_ivars if coord in coords]
            )
        return fields



#!/usr/bin/env python
# coding=utf-8
"""This module regroups all the implemented temporal schemes.
They are written as callable class which take the model and some control
arguments at the init, and perform a computation step every time they are
called.

The following solvers are implemented:
    * Backward and Forward Euler, Crank-Nicolson method (with the Theta class)
    * Some Rosenbrock Wanner schemes (up to the 6th order) with time controler
    * All the scipy.integrate.ode integrators with the scipy_ode class.
"""

import warnings
from functools import partial, wraps

import numpy as np
import scipy.sparse as sps
from fuzzywuzzy import process
from loguru import logger
from scipy.integrate import solve_ivp
from scipy.interpolate import interp1d
from scipy.linalg import norm
from toolz import memoize

available_temporal_schemes = {}


def get_temporal_scheme(name):
    """get a temporal_scheme by its name

    Arguments:
        name {str} -- temporal_scheme name

    Raises:
        NotImplementedError -- raised if the temporal_scheme is not available.

    Returns:
        TemporalScheme -- the requested temporal_scheme
    """
    try:
        return available_temporal_schemes[name.lower()]
    except KeyError:
        err_msg = "%s temporal_scheme is not registered." % name
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            (suggest, score), = process.extract(
                name, available_temporal_schemes.keys(), limit=1
            )
        if score > 70:
            err_msg += (
                "\n%s is available and seems to be close. "
                "It may be what you are looking for !" % suggest
            )
        err_msg += "\nFull list of available temporal_schemes:\n\t- %s" % (
            "\n\t- ".join(available_temporal_schemes.keys())
        )
        raise NotImplementedError(err_msg)


def register_temporal_scheme(CustomTemporalScheme):
    global available_temporal_schemes
    if TemporalScheme not in CustomTemporalScheme.__mro__:
        raise AttributeError(
            "The provider temporal_scheme should inherit from the "
            "TemporalScheme base class."
        )
    available_temporal_schemes[CustomTemporalScheme.name.lower()] = CustomTemporalScheme
    return CustomTemporalScheme


def null_hook(t, fields):
    return fields


def add_time_stepping(scheme, tol=1e-2, ord=2, m=20, reject_factor=2):
    internal_dt = None

    def one_step(t, fields, dt, hook):
        dt_ = dt
        while True:
            _, fields_ = scheme(t, fields, m * dt_, hook)
            for _ in range(m):
                t, fields = scheme(t, fields, dt_, hook)
            errs = [
                np.linalg.norm(
                    np.atleast_1d(fields[key]) - np.atleast_1d(fields_[key]), ord
                )
                / (m ** 2 - 1)
                for key in fields.data_vars
            ]
            err = max(errs)
            dt_ = np.sqrt(dt ** 2 * tol / err)
            if dt_ < dt / reject_factor:
                continue
            break
        return t, fields, dt_

    @wraps(scheme)
    def adaptatif_scheme(t, fields, dt, hook=null_hook):
        nonlocal internal_dt
        next_step = t + dt
        internal_dt = internal_dt if internal_dt else dt
        while t + internal_dt <= next_step:
            t, fields, internal_dt = one_step(t, fields, internal_dt / m, hook)
        if t < next_step:
            t, fields = scheme(t, fields, next_step - t, hook)
        return t, fields

    return adaptatif_scheme


class TemporalScheme:
    pass


class ROW_general(TemporalScheme):
    """Rosenbrock Wanner class of temporal solvers

    The implementation and the different parameters can be found in
    http://www.digibib.tu-bs.de/?docid=00055262
    """

    @memoize
    def __cache__(self, N):
        Id = sps.eye(N, format="csc")
        return Id

    def __init__(
        self,
        model,
        alpha,
        gamma,
        b,
        b_pred=None,
        time_stepping=False,
        tol=None,
        max_iter=None,
        dt_min=None,
        safety_factor=0.9,
        solver="auto",
        recompute_target=True,
        iteratif_atol=1e-3,
        initial_dt=None,
    ):
        self._internal_dt = 1e-6 if initial_dt is None else initial_dt
        self._model = model
        self._alpha = alpha
        self._gamma = gamma
        self._b = b
        self._b_pred = b_pred
        self._s = len(b)
        self._time_control = time_stepping
        self._internal_iter = None
        self._tol = tol
        self._safety_factor = safety_factor
        self._max_iter = max_iter
        self._dt_min = dt_min
        self._recompute_target = recompute_target
        self._solver = solver
        self._interp_cache = None
        self._atol = iteratif_atol

    def __call__(self, t, fields, dt, hook=null_hook):
        """Perform a step of the solver: took a time and a system state as a
          skfdiff Fields container and return the next time step with updated
          container.

          Parameters
          ----------
          t : float
              actual time step
          fields : skfdiff.Fields
              actual system state in a skfdiff Fields
          dt : float
              temporal step-size
          pars : dict
              physical parameters of the model
          hook : callable, optional
              any callable taking the actual time, fields and parameters and
              return modified fields and parameters. Will be called every
              internal time step and can be used to include time dependent or
              conditionnal parameters, boundary conditions...
          container

          Returns
          -------
          tuple : t, fields
              updated time and fields container

          Raises
          ------
          NotImplementedError
              raised if a time stepping is requested but the scheme do not
              provide the b predictor coefficients.
          ValueError
              raised if time_stepping is True and tol is not provided.
          """  # noqa
        if self._time_control:
            return self._variable_step(t, fields, dt, hook=hook)

        t, fields, _ = self._fixed_step(t, fields, dt, hook=hook)
        fields = hook(t, fields)
        return t, fields

    def _fixed_step(self, t, fields, dt, hook=null_hook):
        fields = fields.copy()
        fields = hook(t, fields)
        J = self._model.J(fields, t)
        if self._solver == "auto":
            self._solver = "iteratif" if J.size > 1e4 else "direct"
        U = self._model.U_from_fields(fields)
        Id = self.__cache__(U.size)
        self._A = A = Id - self._gamma[0, 0] * dt * J
        if self._solver == "direct":
            luf = sps.linalg.factorized(A)
        dtF = (
            self._model.F(hook(t + dt, fields), t + dt) - self._model.F(fields, t)
        ) / dt
        ks = []
        for i in np.arange(self._s):
            alpha_i = sum([self._alpha[i, j] for j in range(i)])
            gamma_i = sum([self._gamma[i, j] for j in range(i)])
            t_i = t + alpha_i * dt
            fields_i = hook(
                t_i,
                self._model.fields_from_U(
                    U + dt * sum([self._alpha[i, j] * ks[j] for j in range(i)]), fields
                ),
            )
            F = self._model.F(fields_i, t_i)
            B = (
                F
                + dt
                * (
                    J @ sum([self._gamma[i, j] * ks[j] for j in range(i)])
                    if i > 0
                    else 0
                )
                + dt * gamma_i * dtF
            )
            if self._solver == "direct":
                Utilde = luf(B)
            elif self._solver == "iteratif":
                Utilde = sps.linalg.gcrotmk(
                    A, B, U if not ks else ks[-1], atol=self._atol
                )[0]
            else:
                Utilde = self._solver(A, B)
            ks.append(Utilde)
        U = U + dt * sum([bi * ki for bi, ki in zip(self._b, ks)])

        U_pred = (
            U + dt * sum([bi * ki for bi, ki in zip(self._b_pred, ks)])
            if self._b_pred is not None
            else None
        )
        fields = hook(t + dt, self._model.fields_from_U(U, fields))

        err = norm(U - U_pred, np.inf) if U_pred is not None else None
        return t + dt, fields, err

    def _iter_until_reaching_tol(self, fields, t, dt, hook):
        while self._err is None or self._err > self._tol:
            new_t, new_fields, self._err = self._fixed_step(t, fields, dt, hook)
            logger.debug("error: {}".format(self._err))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                computed_dt = self._safety_factor * dt * np.sqrt(self._tol / self._err)
            if np.isinf(computed_dt):
                computed_dt = 1 / self._safety_factor * dt
            dt = self._internal_dt = computed_dt
        return new_t, dt, new_fields

    def _variable_step(self, t, fields, dt, hook=null_hook):

        self._next_time_step = t + dt
        self._max_dt = t
        self._internal_iter = 0
        try:
            fields = self._model.fields_from_U(
                self._interp_cache(self._next_time_step), fields
            )
            return self._next_time_step, fields
        except (TypeError, ValueError):
            pass
        if not self._recompute_target:
            dt = self._internal_dt
        else:
            dt = self._internal_dt = min(self._internal_dt, dt)
        while True:
            self._err = None
            new_t, dt, new_fields = self._iter_until_reaching_tol(fields, t, dt, hook)

            logger.debug("dt computed after err below tol: %g" % dt)
            logger.debug("ROS_vart, t %g" % t)
            if new_t >= self._next_time_step:
                target_dt = self._next_time_step - t
                logger.debug(
                    "new t more than next expected time step, "
                    "compute with proper timestep %g" % target_dt
                )
                if self._recompute_target:
                    t, fields, self._err = self._fixed_step(
                        t, fields, self._next_time_step - t, hook
                    )
                else:
                    self._interp_cache = interp1d(
                        [t, new_t],
                        [
                            self._model.U_from_fields(fields)[None],
                            self._model.U_from_fields(new_fields)[None],
                        ],
                        axis=0,
                    )
                    fields = self._model.fields_from_U(
                        self._interp_cache(self._next_time_step), fields
                    )
                self._internal_iter += 1
                fields = hook(t, fields)
                return self._next_time_step, fields
            t = new_t
            fields = new_fields.copy()
            self._internal_iter += 1
            if self._internal_iter > (
                self._max_iter if self._max_iter else self._internal_iter + 1
            ):
                raise RuntimeError(
                    "Rosenbrock internal iteration " "above max iterations authorized"
                )
            if dt < (self._dt_min if self._dt_min else dt * 0.5):
                raise RuntimeError(
                    "Rosenbrock internal time step " "less than authorized"
                )


@register_temporal_scheme
class ROS2(ROW_general):
    """Second order Rosenbrock scheme, without time stepping

    Parameters
    ----------
    model : skfdiff.Model
        skfdiff Model
    """

    name = "ROS2"
    embeded_timestepping = False

    def __init__(self, model):
        gamma = np.array(
            [[2.928932188134e-1, 0], [-5.857864376269e-1, 2.928932188134e-1]]
        )
        alpha = np.array([[0, 0], [1, 0]])
        b = np.array([1 / 2, 1 / 2])
        super().__init__(model, alpha, gamma, b, time_stepping=False)


@register_temporal_scheme
class ROS3PRw(ROW_general):
    """Third order Rosenbrock scheme, with time stepping

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      tol : float, optional, default 1E-2
          tolerance factor for the time stepping. The time step will adapt to
          ensure that the maximum relative error on all fields stay under that
          value.
      time_stepping : bool, optional, default True
          allow a variable internal time-step to ensure good agreement between
          computing performance and accuracy.
      max_iter : float or None, optional, default None
          maximum internal iteration allowed
      dt_min : float or None, optional, default None
          minimum internal time step allowed
      recompute_target : bool, optional, default False
          if True a new computation is done when the target time is exceeded,
          interpolation used otherwise.
      """  # noqa

    name = "ROS3PRw"
    embeded_timestepping = True

    def __init__(
        self,
        model,
        tol=1e-1,
        time_stepping=True,
        max_iter=None,
        dt_min=None,
        safety_factor=0.9,
        solver="auto",
        recompute_target=True,
        iteratif_atol=1e-3,
    ):
        alpha = np.zeros((3, 3))
        gamma = np.zeros((3, 3))
        gamma_i = 7.8867513459481287e-01
        b = [5.0544867840851759e-01, -1.1571687603637559e-01, 6.1026819762785800e-01]
        b_pred = [
            2.8973180237214197e-01,
            1.0000000000000001e-01,
            6.1026819762785800e-01,
        ]

        alpha[1, 0] = 2.3660254037844388e00
        alpha[2, 0] = 5.0000000000000000e-01
        alpha[2, 1] = 7.6794919243112270e-01
        gamma[0, 0] = gamma[1, 1] = gamma[2, 2] = gamma_i
        gamma[1, 0] = -2.3660254037844388e00
        gamma[2, 0] = -8.6791218280355165e-01
        gamma[2, 1] = -8.7306695894642317e-01
        super().__init__(
            model,
            alpha,
            gamma,
            b,
            tol=tol,
            b_pred=b_pred,
            time_stepping=time_stepping,
            max_iter=max_iter,
            dt_min=dt_min,
            safety_factor=safety_factor,
            solver=solver,
            recompute_target=recompute_target,
            iteratif_atol=iteratif_atol,
        )


@register_temporal_scheme
class ROS3PRL(ROW_general):
    """4th order Rosenbrock scheme, with time stepping

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      tol : float, optional, default 1E-2
          tolerance factor for the time stepping. The time step will adapt to
          ensure that the maximum relative error on all fields stay under that
          value.
      time_stepping : bool, optional, default True
          allow a variable internal time-step to ensure good agreement between
          computing performance and accuracy.
      max_iter : float or None, optional, default None
          maximum internal iteration allowed
      dt_min : float or None, optional, default None
          minimum internal time step allowed
      recompute_target : bool, optional, default False
          if True a new computation is done when the target time is exceeded,
          interpolation used otherwise.
      """  # noqa

    name = "ROS3PRL"
    embeded_timestepping = True

    def __init__(
        self,
        model,
        tol=1e-1,
        time_stepping=True,
        max_iter=None,
        dt_min=None,
        safety_factor=0.9,
        solver="auto",
        recompute_target=True,
        iteratif_atol=1e-3,
    ):
        alpha = np.zeros((4, 4))
        gamma = np.zeros((4, 4))
        gamma_i = 4.3586652150845900e-01

        b = [
            2.1103008548132443e-03,
            8.8607515441580453e-01,
            -3.2405197677907682e-01,
            4.3586652150845900e-01,
        ]
        b_pred = [
            5.0000000000000000e-01,
            3.8752422953298199e-01,
            -2.0949226315045236e-01,
            3.2196803361747034e-01,
        ]
        alpha[1, 0] = 0.5
        alpha[2, 0] = 0.5
        alpha[2, 1] = 0.5
        alpha[3, 0] = 0.5
        alpha[3, 1] = 0.5
        alpha[3, 2] = 0
        for i in range(len(b)):
            gamma[i, i] = gamma_i
        gamma[1, 0] = -5.0000000000000000e-01
        gamma[2, 0] = -7.9156480420464204e-01
        gamma[2, 1] = 3.5244216792751432e-01
        gamma[3, 0] = -4.9788969914518677e-01
        gamma[3, 1] = 3.8607515441580453e-01
        gamma[3, 2] = -3.2405197677907682e-01
        super().__init__(
            model,
            alpha,
            gamma,
            b,
            b_pred=b_pred,
            tol=tol,
            time_stepping=time_stepping,
            max_iter=max_iter,
            dt_min=dt_min,
            safety_factor=safety_factor,
            solver=solver,
            recompute_target=recompute_target,
            iteratif_atol=iteratif_atol,
        )


@register_temporal_scheme
class RODASPR(ROW_general):
    """6th order Rosenbrock scheme, with time stepping

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      tol : float, optional, default 1E-2
          tolerance factor for the time stepping. The time step will adapt to
          ensure that the maximum relative error on all fields stay under that
          value.
      time_stepping : bool, optional, default True
          allow a variable internal time-step to ensure good agreement between
          computing performance and accuracy.
      max_iter : float or None, optional, default None
          maximum internal iteration allowed
      dt_min : float or None, optional, default None
          minimum internal time step allowed
      recompute_target : bool, optional, default False
          if True a new computation is done when the target time is exceeded,
          interpolation used otherwise.
      """  # noqa

    name = "RODASPR"
    embeded_timestepping = True

    def __init__(
        self,
        model,
        tol=1e-1,
        time_stepping=True,
        max_iter=None,
        dt_min=None,
        safety_factor=0.9,
        solver="auto",
        recompute_target=True,
        iteratif_atol=1e-3,
    ):
        alpha = np.zeros((6, 6))
        gamma = np.zeros((6, 6))
        b = [
            -7.9683251690137014e-1,
            6.2136401428192344e-2,
            1.1198553514719862e00,
            4.7198362114404874e-1,
            -1.0714285714285714e-1,
            2.5e-1,
        ]
        b_pred = [
            -7.3844531665375115e0,
            -3.0593419030174646e-1,
            7.8622074209377981e0,
            5.7817993590145966e-1,
            2.5e-1,
            0,
        ]
        alpha[1, 0] = 7.5e-1
        alpha[2, 0] = 7.5162877593868457e-2
        alpha[2, 1] = 2.4837122406131545e-2
        alpha[3, 0] = 1.6532708886396510e0
        alpha[3, 1] = 2.1545706385445562e-1
        alpha[3, 2] = -1.3157488872766792e0
        alpha[4, 0] = 1.9385003738039885e1
        alpha[4, 1] = 1.2007117225835324e0
        alpha[4, 2] = -1.9337924059522791e1
        alpha[4, 3] = -2.4779140110062559e-1
        alpha[5, 0] = -7.3844531665375115e0
        alpha[5, 1] = -3.0593419030174646e-1
        alpha[5, 2] = 7.8622074209377981e0
        alpha[5, 3] = 5.7817993590145966e-1
        alpha[5, 4] = 2.5e-1
        gamma_i = 0.25
        for i in range(len(b)):
            gamma[i, i] = gamma_i
        gamma[1, 0] = -7.5e-1
        gamma[2, 0] = -8.8644e-2
        gamma[2, 1] = -2.868897e-2
        gamma[3, 0] = -4.84700e0
        gamma[3, 1] = -3.1583e-1
        gamma[3, 2] = 4.9536568e0
        gamma[4, 0] = -2.67694569e1
        gamma[4, 1] = -1.5066459e0
        gamma[4, 2] = 2.720013e1
        gamma[4, 3] = 8.25971337e-1
        gamma[5, 0] = 6.58762e0
        gamma[5, 1] = 3.6807059e-1
        gamma[5, 2] = -6.74235e0
        gamma[5, 3] = -1.061963e-1
        gamma[5, 4] = -3.57142857e-1
        super().__init__(
            model,
            alpha,
            gamma,
            b,
            b_pred=b_pred,
            tol=tol,
            time_stepping=time_stepping,
            max_iter=max_iter,
            dt_min=dt_min,
            safety_factor=safety_factor,
            solver=solver,
            recompute_target=recompute_target,
            iteratif_atol=iteratif_atol,
        )


@register_temporal_scheme
class scipy_ivp(TemporalScheme):
    """Proxy written around the scipy.integrate.solve_ivp function.
      Give access to all
      the scipy integrators.

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      method : str, optional, default 'RK45'
          name of the chosen scipy integration scheme.
      **integrator_kwargs
          extra arguments provided to the scipy integration scheme.
      """  # noqa

    name = "scipy_ivp"
    embeded_timestepping = True

    def __init__(self, model, use_jac=True, method="RK45", **integrator_kwargs):
        self._model = model
        self._method = method
        try:
            model.J
        except AttributeError:
            if method not in ["RK45", "RK23"]:
                warnings.warn(
                    "Jacobian computation routine not available. "
                    "Scipy will try to estimate it, it can be really expensive. "
                    "It's advised to switch to an explicite method, or to change "
                    "the backend."
                )
            use_jac = False
        self._use_jac = use_jac if method not in ["RK45", "RK23"] else False
        self._integrator_kwargs = integrator_kwargs

        def func_scipy_proxy(t, U, fields, hook):
            fields = self._model.fields_from_U(U, fields)
            fields = hook(t, fields)
            return self._model.F(fields, t)

        def jacob_scipy_proxy(t, U, fields, hook):
            fields = self._model.fields_from_U(U, fields)
            fields = hook(t, fields)
            return self._model.J(fields, t)

        self.func = func_scipy_proxy
        self.jac = jacob_scipy_proxy

    def __call__(self, t, fields, dt, hook=null_hook):
        """Perform a step of the solver: took a time and a system state as a
          skfdiff Fields container and return the next time step with updated
          container.

          Parameters
          ----------
          t : float
              actual time step
          fields : skfdiff.Fields
              actual system state in a skfdiff Fields
          dt : float
              temporal step-size
          pars : dict
              physical parameters of the model
          hook : callable, optional
              any callable taking the actual time, fields and parameters and
              return modified fields and parameters. Will be called every
              internal time step and can be used to include time dependent or
              conditionnal parameters, boundary conditions...
          container

          Returns
          -------
          tuple : t, fields
              updated time and fields container

          Raises
          ------
          RuntimeError
              Description
          """  # noqa

        fields = hook(t, fields)
        U = self._model.U_from_fields(fields)

        if self._use_jac:
            self._integrator_kwargs["jac"] = partial(self.jac, fields=fields, hook=hook)

        results = solve_ivp(
            fun=partial(self.func, fields=fields, hook=hook),
            t_span=[t, t + dt],
            t_eval=[t + dt],
            y0=U,
            method=self._method,
            **self._integrator_kwargs
        )
        fields = self._model.fields_from_U(results.y, fields)
        fields = hook(t + dt, fields)
        return t + dt, fields


@register_temporal_scheme
class Theta(TemporalScheme):
    """Simple theta-based scheme where theta is a weight
          if theta = 0, the scheme is a forward-euler scheme
          if theta = 1, the scheme is a backward-euler scheme
          if theta = 0.5, the scheme is called a Crank-Nicolson scheme

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      theta : int, optional, default 1
          weight of the theta-scheme
      solver : callable, optional, default scipy.sparse.linalg.spsolve
          method able to solve a Ax = b linear equation with A a sparse matrix.
          Take A and b as argument and return x.
      """  # noqa

    name = "theta"
    embeded_timestepping = False

    def __init__(self, model, theta=1, solver=sps.linalg.spsolve):
        self._model = model
        self._theta = theta
        self._solver = solver

    def __call__(self, t, fields, dt, hook=null_hook):
        """Perform a step of the solver: took a time and a system state as a
          skfdiff Fields container and return the next time step with updated
          container.

          Parameters
          ----------
          t : float
              actual time step
          fields : skfdiff.Fields
              actual system state in a skfdiff Fields container
          dt : float
              temporal step-size
          pars : dict
              physical parameters of the model
          hook : callable, optional
              any callable taking the actual time, fields and parameters and
              return modified fields and parameters. Will be called every
              internaltime step and can be used to include time dependent or
              conditionnal parameters, boundary conditions...

          Returns
          -------
          tuple : t, fields
              updated time and fields container
          """  # noqa

        fields = fields.copy()
        fields = hook(t, fields)
        U = self._model.U_from_fields(fields)
        F = self._model.F(fields, t)
        if self._theta != 0:
            J = self._model.J(fields, t)
            B = dt * (F - self._theta * J @ U) + U
            J = sps.identity(U.size, format="csc") - self._theta * dt * J
            new_U = self._solver(J, B)
        else:
            new_U = dt * F + U

        if isinstance(new_U, tuple):
            new_U = new_U[0]

        fields = self._model.fields_from_U(new_U, fields)
        fields = hook(t + dt, fields)
        return t + dt, fields


class Stationnary:
    """Solve the stationnary problem F(U) = 0

      Parameters
      ----------
      model : skfdiff.Model
          skfdiff Model
      solver : callable, optional, default scipy.sparse.linalg.spsolve
          method able to solve a Ax = b linear equation with A a sparse matrix.
          Take A and b as argument and return x.
      """  # noqa

    def __init__(self, model, solver=sps.linalg.spsolve):
        self._model = model
        self._solver = solver

    def __call__(self, fields):
        """Perform a step of the solver: took a time and a system state as a
          skfdiff Fields container and return the next time step with updated
          container.

          Parameters
          ----------
          t : float
              actual time step
          fields : skfdiff.Fields
              actual system state in a skfdiff Fields container
          dt : float
              temporal step-size
          pars : dict
              physical parameters of the model
          hook : callable, optional
              any callable taking the actual time, fields and parameters and
              return modified fields and parameters. Will be called every
              internaltime step and can be used to include time dependent or
              conditionnal parameters, boundary conditions...

          Returns
          -------
          tuple : t, fields
              updated time and fields container
          """  # noqa

        fields = fields.copy()
        U0 = self._model.U_from_fields(fields)
        J0 = self._model.J(fields)
        F0 = self._model.F(fields)
        new_U = self._solver(J0, J0 @ U0 - F0)
        if isinstance(new_U, tuple):
            new_U = new_U[0]
        fields = self._model.fields_from_U(new_U, fields)
        return fields



#!/usr/bin/env python
# coding=utf-8

import json
import logging
import warnings
from functools import wraps
import uuid
from collections import deque, namedtuple
from typing import Union
import threading

import yaml
from loguru import logger
from path import Path
from streamz import collect
from xarray import concat, open_dataset, open_mfdataset

FieldsData = namedtuple("FieldsData", ["data", "metadata"])


@wraps(open_dataset)
def _safe_open_dataset(*args, **kwargs):
    with open_dataset(*args, **kwargs) as ds:
        data = ds.copy()
    return data


@wraps(open_mfdataset)
def _safe_open_mfdataset(*args, **kwargs):
    with open_mfdataset(*args, **kwargs) as ds:
        data = ds.compute().copy()
    return data


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def coerce_attr(key, value):
    value_type = type(value)
    if value_type in [int, float, str]:
        return value
    for cast in (int, float, str):
        try:
            value = cast(value)
            logger.debug(
                "Illegal netCDF type ({}) of attribute for {}, "
                "casted to {}".format(value_type, key, cast)
            )
            return value
        except TypeError:
            pass
    raise TypeError(
        "Illegal netCDF type ({}) of attribute for {}, "
        "auto-casting failed, tried to cast to "
        "int, float and str"
    )


class Container:
    def __init__(
        self,
        path=None,
        mode="a",
        *,
        save="all",
        metadata={},
        force=False,
        nbuffer=None,
        save_interval=None
    ):
        if save_interval is None and nbuffer is None:
            save_interval = 10
        if save_interval is not None and nbuffer is not None:
            raise ValueError("You should provide either nbuffer or save_interval.")
        self.nbuffer = nbuffer
        self.save_interval = save_interval
        self._mode = mode
        self._metadata = metadata
        self.save = save
        self._cached_data = deque([], self._n_save)
        self._collector = None
        self._count_iter = 0
        self.path = path = Path(path).abspath() if path else None
        self._writer_thread_name = "writer_%s" % uuid.uuid1()

        if not path:
            return

        if self._mode == "w" and force:
            path.rmtree_p()

        if self._mode == "w" and not force and path.exists():
            raise FileExistsError(
                "Directory %s exists, set force=True to override it" % path
            )

        if self._mode == "r" and not path.exists():
            raise FileNotFoundError("Container not found.")
        path.makedirs_p()

        with open(self.path / "metadata.yml", "w") as yaml_file:
            yaml.dump(self._metadata, yaml_file, default_flow_style=False)

    @property
    def save(self):
        return "last" if self._n_save else "all"

    @save.setter
    def save(self, value):
        if value == "all":
            self._n_save = None
        elif value == "last" or value == -1:
            self._n_save = 1
        else:
            raise ValueError(
                'save argument accept only "all", "last" or -1 '
                "as value, not %s" % value
            )

    def _expand_fields(self, t, fields):
        fields = fields.assign_coords(t=t).expand_dims("t")
        for key, value in self._metadata.items():
            fields.attrs[key] = coerce_attr(key, value)
        self._cached_data.append(fields)
        return fields

    def _concat_fields(self, fields):
        if fields:
            return concat(fields, dim="t")

    def connect(self, stream):
        def get_t_fields(simul):
            return simul.t, simul.fields

        def expand_fields(inps):
            return self._expand_fields(*inps)

        def get_last(list_fields):
            try:
                return list_fields[-1]
            except IndexError:
                pass

        accumulation_stream = stream.map(get_t_fields).map(expand_fields)
        self._collector = collect(accumulation_stream)
        if self.save == "all":
            self._collector.map(self._concat_fields).sink(self._bg_write)
        else:
            self._collector.map(get_last).sink(self._bg_write)

        if self.nbuffer is not None:
            (accumulation_stream.partition(self.nbuffer).sink(self._collector.flush))
        if self.save_interval is not None:
            (
                accumulation_stream.timed_window(self.save_interval)
                .filter(bool)
                .sink(self._collector.flush)
            )

        return self._collector

    @property
    def writers(self):
        return [
            thread
            for thread in threading.enumerate()
            if thread.name == self._writer_thread_name
        ]

    @property
    def is_writing(self):
        if self.writers:
            return True
        return False

    def flush(self):
        if self._collector:
            self._collector.flush()
            while self.is_writing:
                pass

    def _write(self, concatenated_fields):
        if concatenated_fields is not None and self.path:
            len_concat_fields = concatenated_fields.t.size
            target_file = (
                self.path
                / "data_%i-%i.nc"
                % (self._count_iter, self._count_iter + len_concat_fields - 1)
            )
            self._count_iter = len_concat_fields + self._count_iter
            concatenated_fields.to_netcdf(target_file)
            concatenated_fields.close()
            self._cached_data = deque([], self._n_save)
            if self.save == "last":
                [
                    file.remove()
                    for file in self.path.glob("data_*.nc")
                    if file != target_file
                ]

    def _bg_write(self, concatenated_fields):
        thread = threading.Thread(
            name=self._writer_thread_name,
            target=self._write,
            args=(concatenated_fields,),
        )
        thread.start()

    def __repr__(self):
        repr = """path:   {path}
{data}""".format(
            path=self.path, data=self.data if self.data is not None else "Empty"
        )
        return repr

    def __del__(self):
        self.flush()

    @property
    def data(self):
        try:
            self.flush()
            if self.path:
                return _safe_open_mfdataset(self.path / "data*.nc")
            return self._concat_fields(self._cached_data)
        except OSError:
            return

    @property
    def metadata(self):
        try:
            if self.path:
                with open(self.path / "metadata.yml", "r") as yaml_file:
                    return yaml.load(yaml_file)
            return self._metadata
        except OSError:
            return

    @metadata.setter
    def metadata(self, parameters):
        if self._mode == "r":
            return
        for key, value in parameters.items():
            self._metadata[key] = value
        if self.path:
            with open(self.path / "metadata.yml", "w") as yaml_file:
                yaml.dump(self._metadata, yaml_file, default_flow_style=False)

    @staticmethod
    def retrieve(path: Path, isel: Union[str, dict, int] = "all", lazy: bool = False):
        """Retrieve the data of a persistent container.

        Parameters
        ----------
        path: Path or str
            The folder where the persistent container lives.
        isel : Union[str, dict, int], optional
            can be either "all" or "last", an integer or a sequence of integer.
            (the default is "all")
        lazy : bool, optional
            if True, return a lazy xarray Datasets that will be loaded when requested
            by the user (using the :py:func:`compute` method). Useful when the data
            are too big to fit in memory. (the default is False)

        Returns
        -------
        :py:class:`FieldsData`
            The requested data, with :py:attr:`FieldsData.data` the fields dataset and
            :py:attr:`FieldsData.metadata` the metadata dictionary.
        """
        path = Path(path)
        if isel == "last":
            last_file = sorted(
                [filename for filename in path.files("data*.nc")],
                key=lambda filename: int(
                    filename.basename().stripext().split("_")[-1].split("-")[-1]
                ),
            )[-1]
            data = _safe_open_dataset(last_file).isel(t=-1)
        else:
            if lazy:
                data = _safe_open_mfdataset(path / "data*.nc", concat_dim="t").sortby(
                    "t"
                )
            else:
                data = concat(
                    [open_dataset(filename) for filename in path.files("data*.nc")],
                    dim="t",
                ).sortby("t")

        with open(Path(path) / "metadata.yml", "r") as yaml_file:
            metadata = yaml.load(yaml_file)

        if isel not in ["all", "last"]:
            data = data.isel(t=isel)

        return FieldsData(data=data, metadata=AttrDict(**metadata))

    def merge(self, override=True):
        if self.path:
            return Container.merge_datafiles(self.path, override=override)

    @staticmethod
    def merge_datafiles(path, override=False):
        path = Path(path)

        if (path / "data.nc").exists() and not override:
            raise FileExistsError(path / "merged_data.nc")
        (path / "data.nc").remove_p()

        split_data = _safe_open_mfdataset(path / "data*.nc", concat_dim="t").sortby("t")
        split_data.to_netcdf(path / "merged_data.nc")
        merged_data = _safe_open_dataset(
            path / "merged_data.nc", chunks=split_data.chunks
        )

        if not split_data.equals(merged_data):
            (path / "merged_data.nc").remove()
            raise IOError("Unable to merge data ")

        split_data.close()
        merged_data.close()

        return path / "merged_data.nc"



#!/usr/bin/env python
# coding=utf-8

from functools import partial
from operator import attrgetter

import attr
import numpy as np
from loguru import logger
from scipy.sparse import csc_matrix
from sympy import lambdify, Symbol
from cachetools import cachedmethod

from .base_backend import Backend, register_backend


def np_depvar_printer(unk):
    return Symbol(str(unk.discrete))


def np_ivar_printer(coord):
    return (Symbol(str(coord.discrete)), Symbol(str(coord.idx)), coord.step, coord.N)


def np_Min(args):
    a, b = args
    return np.where(a < b, a, b)


def np_Max(args):
    a, b = args
    return np.where(a < b, b, a)


def np_Heaviside(a):
    return np.where(a < 0, 0, 1)


lambdify = partial(
    lambdify,
    modules=[{"amax": np_Max, "amin": np_Min, "Heaviside": np_Heaviside}, "scipy"],
)


@register_backend
@attr.s
class NumpyBackend(Backend):
    name = "numpy"

    def __attrs_post_init__(self):
        logger.info("numpy backend: convert_inputs...")
        self._convert_inputs()
        logger.info("numpy backend: build_evolution_equations...")
        self._build_evolution_equations()
        logger.info("numpy backend: build_jacobian...")
        self._cache_jac_coord = {}
        self._build_jacobian()

    def _convert_inputs(self):
        self.ndim = len(self.system.coordinates)
        self.unks = list(map(np_depvar_printer, self.system.unknowns))
        self.pars = list(map(np_depvar_printer, self.system.parameters))
        self.coords, self.idxs, self.steps, self.sizes = zip(
            *map(np_ivar_printer, self.system.coordinates)
        )
        self.idxs_map = [
            tuple([self.system.coordinates.index(coord) for coord in unk.coordinates])
            for unk in self.system.unknowns
        ]

        self.shapes = [self.system.shapes[unk] for unk in self.system.unknowns]

        self.inputs = [
            self.system._t,
            *self.unks,
            *self.pars,
            *self.coords,
            *self.idxs,
            *self.steps,
            *self.sizes,
        ]
        self.inputs_cond = [*self.idxs, *self.sizes]

    def _build_evolution_equations(self):
        self.evolution_equations = [
            lambdify(self.inputs, expr.n()) for expr in self.system.piecewise_system
        ]

    def compute_F_vector(self, t, unks, pars, coords, sizes):
        subgrids = self.grid_builder.compute_subgrids(sizes)
        steps = self.grid_builder.compute_steps(
            tuple(sizes), tuple([coord.ptp() for coord in coords])
        )
        system_sizes = self.grid_builder.compute_sizes(sizes)
        system_size = sum(system_sizes)
        F = np.empty(system_size)
        for grid, eq in zip(subgrids, self.evolution_equations):
            Fi = eq(t, *unks, *pars, *coords, *grid[:, 1:-2].T, *steps, *sizes)
            F[grid[:, -1]] = Fi
        return F

    def F(self, fields, t=0):
        unks = [
            fields[varname].values
            for varname in [unk.name for unk in self.system.unknowns]
        ]
        pars = [
            fields[varname].values
            for varname in [par.name for par in self.system.parameters]
        ]
        coords = [
            fields[varname].values
            for varname in [coord.name for coord in self.system.coordinates]
        ]
        sizes = [coord.size for coord in coords]

        return self.compute_F_vector(t, unks, pars, coords, sizes)

    def _build_jacobian(self):

        self.jacobian_routines = [
            [lambdify(self.inputs, diff) for diff in jac_row]
            for jac_row in self.system.jacobian_values
        ]
        self.jacobian_columns = [
            [lambdify(self.inputs_cond, col) for col in columns]
            for columns in self.system.jacobian_columns
        ]

    def compute_jacobian_values(self, t, unks, pars, coords, sizes):
        subgrids = self.grid_builder.compute_subgrids(sizes)
        steps = self.grid_builder.compute_steps(
            tuple(sizes), tuple([coord.ptp() for coord in coords])
        )
        data_size = sum(
            [
                subgrid.shape[0] * len(jacs)
                for subgrid, jacs in zip(subgrids, self.jacobian_columns)
            ]
        )
        data = np.zeros(data_size)

        cursor = 0
        for grid, jacs in zip(subgrids, self.jacobian_routines):
            for jac_func in jacs:
                next_cursor = cursor + grid.shape[0]
                jac = jac_func(
                    t, *unks, *pars, *coords, *grid[:, 1:-2].T, *steps, *sizes
                )
                data[cursor:next_cursor] = jac
                cursor = next_cursor
        return data

    @cachedmethod(cache=attrgetter("_cache_jac_coord"))
    def compute_jacobian_coordinates(self, *sizes):
        subgrids = self.grid_builder.compute_subgrids(sizes)
        system_sizes = self.grid_builder.compute_sizes(sizes)
        system_size = sum(system_sizes)

        rows_list = []
        cols_list = []
        for grid, jac_cols in zip(subgrids, self.jacobian_columns):
            for col_func in jac_cols:
                cols_ = np.zeros((grid.shape[0], self.ndim + 1), dtype="int32")
                cols = col_func(*grid[:, 1:-2].T, *sizes)
                cols = np.stack(
                    [np.broadcast_to(col, cols_.shape[:-1]) for col in cols]
                )

                flat_cols = self.grid_builder.compute_flat_from_idxs(cols.T, sizes)

                rows_list.extend(grid[:, -1].reshape((-1,)))
                cols_list.extend(flat_cols.reshape((-1,)))
        rows = np.array(rows_list)
        cols = np.array(cols_list)

        perm = np.argsort(cols)
        perm_rows = rows[perm]
        perm_cols = cols[perm]
        count = np.zeros((system_size + 1), dtype="int32")
        uq, cnt = np.unique(perm_cols, False, False, True)
        count[uq + 1] = cnt
        indptr = np.cumsum(count)
        return rows, cols, perm_rows, indptr, perm, (system_size, system_size)

    def J(self, fields, t=0):
        unks = [
            fields[varname].values
            for varname in [unk.name for unk in self.system.unknowns]
        ]
        pars = [
            fields[varname].values
            for varname in [par.name for par in self.system.parameters]
        ]
        coords = [
            fields[varname].values
            for varname in [coord.name for coord in self.system.coordinates]
        ]
        sizes = [coord.size for coord in coords]

        data = self.compute_jacobian_values(t, unks, pars, coords, sizes)
        _, _, perm_rows, indptr, perm, shape = self.compute_jacobian_coordinates(*sizes)

        return csc_matrix((data[perm], perm_rows, indptr), shape=shape)



#!/usr/bin/env python
# coding=utf-8

from collections import defaultdict
from functools import partial

from holoviews import Curve, DynamicMap, Image, Layout, streams
from loguru import logger
from path import Path  # noqa


class Display:
    def __init__(self, skel_data, plot_function):

        self._plot_pipe = streams.Pipe(data=skel_data)
        self._dynmap = DynamicMap(plot_function, streams=[self._plot_pipe])

    def _repr_mimebundle_(self, *args, **kwargs):
        return self.hv_curve._repr_mimebundle_(*args, **kwargs)

    def connect(self, stream, n=1):
        stream.partition(n).pluck(-1).sink(self._plot_pipe.send)

    @property
    def hv_curve(self):
        return self._dynmap.collate()

    def show(self):
        return self.hv_curve

    def __add__(self, other):
        if isinstance(other, Display):
            return self._dynmap + other._dynmap
        return self._dynmap + other

    def __mul__(self, other):
        if isinstance(other, Display):
            return self._dynmap * other._dynmap
        self._dynmap * other

    @staticmethod
    def display_custom(simul, plot_function, n=1):
        def wrapped_plot_function(data):
            return plot_function(data)

        display = Display(simul, wrapped_plot_function)
        display.connect(simul.stream, n)
        return display

    @staticmethod
    def display_fields(simul, keys="unknowns", n=1, dim_allowed=(0, 1, 2)):
        _0D_history = defaultdict(
            list
        )  # mypy: Dict[str, Sequence[Tuple[float, float]]]

        def plot_function(keys, _0D_history, data):
            curves = []
            if keys == "all":
                keys = data.fields.data_vars
            elif keys == "unknowns":
                parameters = [par.name for par in simul.model.pdesys.parameters]
                keys = [
                    key
                    for key in data.fields.data_vars
                    if key in set(data.fields.data_vars).difference(parameters)
                ]
            if not keys:
                raise ValueError("keys are empty.")
            for var in keys:
                displayed_field = data.fields[var]
                if 0 in dim_allowed and len(displayed_field.dims) == 0:
                    _0D_history[var].append((simul.t, float(data.fields[var])))
                    curve = Curve(_0D_history[var], kdims="t", vdims=var, label=var)
                    curves.append(curve)

                if 1 in dim_allowed and len(displayed_field.dims) == 1:
                    curves.append(Curve((displayed_field.squeeze()), label=var))
                elif 2 in dim_allowed and len(displayed_field.dims) == 2:
                    curves.append(Image((displayed_field.squeeze()), label=var))
                else:
                    continue
            return Layout(curves)

        display = Display(simul, partial(plot_function, keys, _0D_history))
        display.connect(simul.stream, n)
        display.cols = display.hv_curve.cols
        return display



#!/usr/bin/env python
# coding=utf-8

import warnings
from abc import ABC

import attr
from fuzzywuzzy import process
import cloudpickle

from ..grid_builder import GridBuilder
from ..system import PDESys

available_backends = {}


def get_backend(name):
    """get a backend by its name

    Arguments:
        name {str} -- backend name

    Raises:
        NotImplementedError -- raised if the backend is not available.

    Returns:
        Backend -- the requested backend
    """
    try:
        return available_backends[name]
    except KeyError:
        err_msg = "%s backend is not registered." % name
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            (suggest, score), = process.extract(
                name, available_backends.keys(), limit=1
            )
        if score > 70:
            err_msg += (
                "\n%s is available and seems to be close. "
                "It may be what you are looking for !" % suggest
            )
        err_msg += "\nFull list of available backends:\n\t- %s" % (
            "\n\t- ".join(available_backends.keys())
        )
        raise NotImplementedError(err_msg)


def register_backend(CustomBackend):
    global available_backends
    if Backend not in CustomBackend.__mro__:
        raise AttributeError(
            "The provider backend should inherit from the " "Backend base class."
        )
    available_backends[CustomBackend.name] = CustomBackend
    return CustomBackend


@attr.s
class Backend(ABC):
    system = attr.ib(type=PDESys)
    grid_builder = attr.ib(type=GridBuilder)

    def __getstate__(self):
        return {key: cloudpickle.dumps(value) for key, value in self.__dict__.items()}

    def __setstate__(self, state):
        self.__dict__.update(
            {key: cloudpickle.loads(value) for key, value in state.items()}
        )



#!/usr/bin/env python
# coding=utf-8

import warnings
from .base_backend import get_backend, available_backends

try:
    from .numba_backend import NumbaBackend
except ImportError:
    warnings.warn("Numba cannot be imported: numba backend will not be available.")

from .numpy_backend import NumpyBackend



#!/usr/bin/env python
# coding=utf-8

import inspect
import time
import warnings
from collections import namedtuple
from datetime import datetime, timedelta
from uuid import uuid4

import cloudpickle
import streamz
from loguru import logger
from numpy import isclose
import xarray as xr

from .plugins.container import Container
from .plugins.displays import Display
from .utils import tqdm
from .core.temporal_schemes import add_time_stepping, get_temporal_scheme

now = datetime.now


class Timer:
    def __init__(self, last, total):
        self.last = last
        self.total = total

    def __repr__(self):
        repr = """last:   {last}
total:  {total}"""
        return repr.format(
            last=(now() - timedelta(seconds=self.last)),
            total=(now() - timedelta(seconds=self.total)),
        )


def null_hook(t, fields):
    return fields


def get_total_iter(tmax, user_dt):
    if tmax:
        return tmax // user_dt
    return None


def get_initial(total_iter, i):
    if total_iter and i < total_iter:
        return i
    if total_iter:
        return total_iter
    return 0


def postprocess_save_timer(simul):
    simul.fields["last_run_time"] = simul.timer.last
    simul.fields["total_run_time"] = simul.timer.total


PostProcess = namedtuple("PostProcess", ["name", "function", "description"])


def _reduce_simulation(
    model,
    fields,
    dt,
    t,
    tmax,
    i,
    id,
    _pprocesses,
    _scheme,
    status,
    _total_running,
    _last_running,
    _created_timestamp,
    _started_timestamp,
    _last_timestamp,
    _actual_timestamp,
    _hook,
    _container,
):
    simul = Simulation(model=model, fields=fields, dt=dt, t=t, tmax=tmax, id=id)
    simul.i = i
    simul._pprocesses = cloudpickle.loads(_pprocesses)
    simul._scheme = cloudpickle.loads(_scheme)
    simul.status = status
    simul._total_running = _total_running
    simul._last_running = _last_running
    simul._created_timestamp = _created_timestamp
    simul._started_timestamp = _started_timestamp
    simul._last_timestamp = _last_timestamp
    simul._actual_timestamp = _actual_timestamp
    simul._hook = _hook
    simul._container = _container

    return simul


class Simulation(object):
    """High level container used to run simulation build on skfdiff Model.
      This object is an iterable which will yield every time step until the
      parameters 'tmax' is reached if provided.
      By default, the solver use a 6th order ROW solver, an implicit method
      with integrated time-stepping.

      Parameters
      ----------
      model : skfdiff.Model
          Contain finite difference approximation and routine of the dynamical
          system
      fields : skfdiff.BaseFields or dict (any mappable)
          skfdiff container or mappable filled with initial conditions
      dt : float
          time stepping for output. if time_stepping is False, the internal
          time stepping will be the same.
      t : float, optional, default 0.
          initial time
      tmax : float, optional, default None
          Control the end of the simulation. If None (the default), the com-
          putation will continue until interrupted by the user (using Ctrl-C
          or a SIGTERM signal).
      id : None, optional
          Name of the simulation. A 2 word slug will be generated if not
          provided.
      hook : callable, optional, default null_hook.
          Any callable taking the actual time, fields and parameters and
          return modified fields and parameters.
          Will be called every internal time step and can be used to include
          time dependent or conditionnal parameters, boundary conditions...
          The default null_hook has no impact on the computation.
      scheme : callable, optional, default skfdiff.schemes.RODASPR
          An callable object which take the simulation state and return
          the next step.
          Its signature is scheme.__call__(fields, t, dt, hook)
          and it should return the next time and the updated fields.
          It take the model and extra positional and named arguments.
      time_stepping : boolean, default True
          Indicate if the time step is controlled by an algorithm dependant of
          the temporal scheme (see the doc on time stepping for extra info).
      **kwargs
          extra arguments passed to the scheme.

      Attributes
      ----------
      dt : float
        output time step
      fields : skfdiff.Fields
        skfdiff container filled with actual data
      i : int
        actual iteration
      id : str
        name of the simulation
      model : skfdiff.Model
        skfdiff Model used in the simulation
      status : str
        status of the simulation, one of the following one:
        ('created', 'running', 'finished', 'failed')
      t : float
        actual time
      tmax : float or None, default None
        stopping time of the simulation. Not stopping if set to None.

      Examples
      --------
      >>> import numpy as np
      >>> import skfdiff
      >>> model = skfdiff.Model(["k1 * dxxU",
      ...                        "k2 * dxxV"],
      ...                       ["U", "V"],
      ...                       ["k1", "k2"])
      >>> x = np.linspace(0, 100, 1000, endpoint=False)
      >>> U = np.cos(x * 2 * np.pi / 100)
      >>> V = np.sin(x * 2 * np.pi / 100)
      >>> fields = model.Fields(x=x, U=U, V=V)
      >>> simulation = skfdiff.Simulation(model, fields, dt=5., tmax=50.)
      >>> for t, fields in simulation:
      ...    pass
      >>> print(t)
      50.0
      """  # noqa

    def __init__(
        self,
        model,
        fields,
        dt,
        t=0,
        tmax=None,
        id=None,
        hook=null_hook,
        scheme="RODASPR",
        time_stepping=True,
        **kwargs
    ):
        def intersection_kwargs(kwargs, function):
            """Inspect the function signature to identify the relevant keys
            in a dictionary of named parameters.
            """
            func_signature = inspect.signature(function)
            func_parameters = func_signature.parameters
            kwargs = {
                key: value for key, value in kwargs.items() if key in func_parameters
            }
            return kwargs

        kwargs["time_stepping"] = time_stepping
        self.id = str(uuid4())[:6] if not id else id
        self.model = model
        if not isinstance(fields, xr.Dataset):
            fields = model.Fields(**{var: fields[var] for var in fields.keys()})
        self.fields = fields
        self.t = t
        self.user_dt = self.dt = dt
        self.tmax = tmax
        self.i = 0
        self._stream = streamz.Stream()
        self._pprocesses = []
        scheme = get_temporal_scheme(scheme)
        self._scheme = scheme(model, **intersection_kwargs(kwargs, scheme.__init__))
        if time_stepping and not self._scheme.embeded_timestepping:
            self._scheme = add_time_stepping(
                self._scheme, **intersection_kwargs(kwargs, add_time_stepping)
            )

        self.status = "created"

        self._total_running = 0
        self._last_running = 0

        self._total_plugins_running = 0
        self._last_plugins_running = 0

        self._created_timestamp = now()
        self._started_timestamp = None
        self._last_timestamp = None
        self._actual_timestamp = now()
        self._hook = hook
        self._container = None
        self._iterator = self.compute()

    def _compute_one_step(self, t, fields):
        """
        Compute one step of the simulation, then update the timers.
        """
        fields = self._hook(t, fields)
        self.dt = self.tmax - t if self.tmax and (t + self.dt >= self.tmax) else self.dt
        before_compute = time.clock()
        t, fields = self._scheme(t, fields, self.dt, hook=self._hook)
        after_compute = time.clock()
        self._last_running = after_compute - before_compute
        self._total_running += self._last_running
        self._last_timestamp = self._actual_timestamp
        self._actual_timestamp = now()
        return t, fields

    def compute(self):
        """Generator which yield the actual state of the system every dt.

        Yields
        ------
        tuple : t, fields
            Actual time and updated fields container.
        """
        fields = self.fields
        t = self.t
        self._started_timestamp = now()
        self.stream.emit(self)

        if self.tmax and isclose(self.t, self.tmax):
            self._end_simulation()
            return

        try:
            while True:
                t, fields = self._compute_one_step(t, fields)

                self.i += 1
                self.t = t
                self.fields = fields
                for pprocess in self.post_processes:
                    pprocess.function(self)

                before_emit = time.time()
                self.stream.emit(self)
                after_emit = time.time()
                self._last_plugins_running = after_emit - before_emit
                self._total_plugins_running += self._last_plugins_running

                yield self.t, self.fields

                if self.tmax and isclose(self.t, self.tmax):
                    self._end_simulation()
                    return

        except RuntimeError:
            self.status = "failed"
            raise

    def _end_simulation(self):
        if self.container:
            self.container.flush()

    def _run_with_progress(self, total_iter, initial, log):
        with tqdm(initial=initial, total=total_iter) as pbar:
            for t, fields in self:
                pbar.update(1)
                pbar.set_description("%s running: t: %g" % (self.id, t))
                log("%s running: t: %g" % (self.id, t))
            try:
                return t, fields
            except UnboundLocalError:
                warnings.warn("Simulation already ended")

    def _run_without_progress(self, log):
        for t, fields in self:
            log("%s running: t: %g" % (self.id, t))
        try:
            return t, fields
        except UnboundLocalError:
            warnings.warn("Simulation already ended")

    def run(self, progress=True, verbose=False):
        """Compute all steps of the simulation. Be careful: if tmax is not set,
        this function will result in an infinit loop.

        Returns
        -------

        (t, fields):
            last time and result fields.
        """

        log = logger.info if verbose else logger.debug

        total_iter = get_total_iter(self.tmax, self.user_dt)
        initial = get_initial(total_iter, self.i)

        if progress:
            return self._run_with_progress(total_iter, initial, log)
        return self._run_without_progress(log)

    def __repr__(self):
        repr = """{simulation_name:=^30}

created:      {created_date}
started:      {started_date}
last:         {last_date}

time:         {t:g}
iteration:    {iter:g}

last step:    {step_time}
total time:   {running_time}

Hook function
-------------
{hook_source}

Container
---------
{container}

=========== Model ===========
{model_repr}"""
        repr = repr.format(
            simulation_name=" %s " % self.id,
            container=self.container,
            t=self.t,
            iter=self.i,
            model_repr=self.model,
            hook_source=inspect.getsource(self._hook),
            step_time=self._last_running,
            running_time=self._total_running,
            created_date=(self._created_timestamp),
            started_date=(
                self._started_timestamp if self._started_timestamp else "None"
            ),
            last_date=(self._last_timestamp if self._last_timestamp else "None"),
        )
        return repr

    def attach_container(
        self,
        path=None,
        save="all",
        mode="w",
        nbuffer=None,
        save_interval=None,
        force=False,
    ):
        """add a Container to the simulation which allows some
        persistance to the simulation.

        Parameters
        ----------
        path : str or None (default: None)
            path for the container. If None (the default), the data lives only
            in memory (and are available with `simulation.container`)
        mode : str, optional
            "a" or "w" (default "w")
        save : str, optional
            "all" will save every time-step,
            "last" will only get the last time step
        nbuffer : int, optional
            wait until nbuffer data in the Queue before saving on disk.
        save_interval : int, optional
            wait until save_interval since last flush before saving on disk.
        force : bool, optional (default False)
            if True, remove the target folder if not empty. if False, raise an
            error.
        """
        self._container = Container(
            "%s/%s" % (path, self.id) if path else None,
            save=save,
            mode=mode,
            force=force,
            nbuffer=nbuffer,
            save_interval=save_interval,
        )
        self._container.connect(self.stream)
        return self._container

    @property
    def post_processes(self):
        """list of skfdiff.core.simulation.PostProcess:
        contain all the post processing function attached to the simulation.
        """
        return self._pprocesses

    @property
    def stream(self):
        """ streamz.Stream: Streamz starting point, fed by the simulation state after
        each time_step. This interface is used for post-processing, saving the data on
        disk by the Container and display the fields in real-time.
        """
        return self._stream

    @property
    def container(self):
        """skfdiff.Container: give access to the attached container, if any.
        """
        return self._container

    def display(self, keys="unknowns", n=1, dim_allowed=(0, 1, 2)):
        return Display.display_fields(self, keys=keys, n=n, dim_allowed=dim_allowed)

    def display_custom(self, plot_function, n=1):
        return Display.display_custom(self, plot_function, n=n)

    @property
    def timer(self):
        """ skfdiff.core.simulation.Timer: the cpu time of the previous step and
        the total running time of the simulation.
        """
        return Timer(self._last_running, self._total_running)

    @property
    def plugins_timer(self):
        """ skfdiff.core.simulation.Timer: the cpu time of the plugins (post process,
        data save, display...).
        """
        return Timer(self._last_plugins_running, self._total_plugins_running)

    def add_post_process(self, name, post_process, description=""):
        """add a post-process

        Parameters
        ----------
        name : str
            name of the post-traitment
        post_process : callback (function of a class with a __call__ method
                                 or a streamz.Stream).
            this callback have to accept the simulation state as parameter
            and return the modifield simulation state.
            if a streamz.Stream is provided, it will me plugged_in with the
            previous streamz (and ultimately to the initial_stream). All these
            stream accept and return the simulation state.
        description : str, optional, Default is "".
            give extra information about the post-processing
        """

        self._pprocesses.append(
            PostProcess(name=name, function=post_process, description=description)
        )
        self._pprocesses[-1].function(self)

    def remove_post_process(self, name):
        """remove a post-process

        Parameters
        ----------
        name : str
            name of the post-process to remove.
        """
        self._pprocesses = [
            post_process
            for post_process in self._pprocesses
            if post_process.name != name
        ]

    def __iter__(self):
        return self.compute()

    def __next__(self):
        return next(self._iterator)

    def __reduce__(self):
        return (
            _reduce_simulation,
            (
                self.model,
                self.fields,
                self.dt,
                self.t,
                self.tmax,
                self.i,
                self.id,
                cloudpickle.dumps(self._pprocesses),
                cloudpickle.dumps(self._scheme),
                self.status,
                self._total_running,
                self._last_running,
                self._created_timestamp,
                self._started_timestamp,
                self._last_timestamp,
                self._actual_timestamp,
                self._hook,
                self._container,
            ),
        )



#!/usr/bin/env python
# coding=utf-8

import re
import string
import typing
from functools import reduce
from itertools import product
from operator import and_

import attr
from sympy import Eq, Function, Idx, IndexedBase, Symbol


@attr.s(frozen=True, repr=False, hash=False)
class Coordinate:
    name = attr.ib(type=str, converter=str)

    @name.validator
    def check(self, attrs, value):
        if not isinstance(value, Coordinate) and (
            value not in string.ascii_letters or len(value) > 1
        ):
            raise ValueError(
                "independant variables have to be 1 char "
                "lenght and be an ascii letter "
                '(is "%s")' % value
            )

    @property
    def coords(self):
        return (self,)

    @property
    def c_symbols(self):
        return tuple([coord.symbol for coord in self.coords])

    @property
    def c_names(self):
        return tuple([coord.name for coord in self.coords])

    @property
    def c_steps(self):
        return tuple([coord.step for coord in self.coords])

    @property
    def c_step_values(self):
        return tuple([coord.step_value for coord in self.coords])

    @property
    def c_discs(self):
        return tuple([coord.discrete for coord in self.coords])

    @property
    def c_idxs(self):
        return tuple([coord.idx for coord in self.coords])

    @property
    def c_Ns(self):
        return tuple([coord.N for coord in self.coords])

    @property
    def c_bounds(self):
        return tuple([coord.bound for coord in self.coords])

    @property
    def symbol(self):
        return Symbol(self.name)

    @property
    def discrete(self):
        return IndexedBase(self.name)

    @property
    def N(self):
        return Symbol("N_%s" % self.name, integer=True)

    @property
    def bound(self):
        return (0, self.N - 1)

    def get_node_coord(self, domain):
        if domain == "left":
            return self.idx.lower
        if domain == "right":
            return self.idx.upper
        return self.idx

    def domains(self, *node_coords):
        return tuple(
            [
                coord.domain(node_coord)
                for coord, node_coord in zip(self.coords, node_coords)
            ]
        )

    def domain(self, node_coord):
        def to_bool(cond):
            try:
                cond = cond.subs(self.N, 500)
            except AttributeError:
                pass
            return cond

        is_left = to_bool(node_coord <= self.idx.lower)
        is_right = to_bool(node_coord >= self.idx.upper)
        # if x_idx is in the node_coord, it belong to the bulk.
        # Otherwise, it's only if it's neither into the left and the right.
        try:
            if str(self.idx) in map(str, node_coord.atoms()):
                return "bulk"
        except AttributeError:
            pass
        if not is_left and not is_right:
            return "bulk"
        if is_left:
            return "left"
        if is_right:
            return "right"

    def is_in_bulk(self, node_coord):
        return True if self.domain(node_coord) == "bulk" else False

    def distance_from_domain(self, node_coord):
        if self.domain(node_coord) == "bulk":
            return 0
        if self.domain(node_coord) == "left":
            return int(self.idx.lower - node_coord)
        if self.domain(node_coord) == "right":
            return int(node_coord - self.idx.upper)

    def distances(self, node_coord):
        return (self.distance_from_domain(node_coord),)

    @property
    def idx(self):
        return Idx("%s_idx" % self.name, self.N)

    @property
    def step(self):
        return Symbol("d%s" % self.name)

    @property
    def step_value(self):
        return (self.discrete[self.idx.upper] - self.discrete[self.idx.lower]) / (
            self.N - 1
        )

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.name)


_dependent_var_re = re.compile(r"^(?P<unk>\w+)(?:\((?P<unkarg>[^\)]*)\))?$")


@attr.s(frozen=True, repr=False, hash=False)
class Unknown:
    name = attr.ib(type=str)
    coordinates = attr.ib(init=False, type=typing.Tuple[Coordinate])

    def __attrs_post_init__(self):
        if isinstance(self.name, Unknown):
            name, coordinates = self.name.name, self.name.coords
        else:
            name, coordinates = Unknown._convert(self.name)
        object.__setattr__(self, "name", name)
        object.__setattr__(self, "coordinates", coordinates)

    @staticmethod
    def _convert(var):
        match = _dependent_var_re.match(var)
        if not match:
            raise ValueError("Dependent variable not properly formatted.")

        unk, coords = _dependent_var_re.findall(var)[0]
        return (
            unk,
            tuple(
                [
                    Coordinate(coord.strip())
                    for coord in coords.split(",")
                    if coord != ""
                ]
            ),
        )

    @property
    def coords(self):
        return self.coordinates

    @property
    def c_symbols(self):
        return tuple([coord.symbol for coord in self.coords])

    @property
    def c_names(self):
        return tuple([coord.name for coord in self.coords])

    @property
    def c_steps(self):
        return tuple([coord.step for coord in self.coords])

    @property
    def c_step_values(self):
        return tuple([coord.step_value for coord in self.coords])

    @property
    def c_discs(self):
        return tuple([coord.discrete for coord in self.coords])

    @property
    def c_idxs(self):
        return tuple([coord.idx for coord in self.coords])

    @property
    def c_Ns(self):
        return tuple([coord.N for coord in self.coords])

    @property
    def c_bounds(self):
        return tuple([coord.bound for coord in self.coords])

    def domains(self, *node_coords):
        return tuple(
            [
                coord.domain(node_coord)
                for coord, node_coord in zip(self.coords, node_coords)
            ]
        )

    def distances(self, *node_coords):
        return [
            coord.distance_from_domain(node_coord)
            for coord, node_coord in zip(self.coords, node_coords)
        ]

    @property
    def physical_domains(self):
        domains = product(*[("left", "bulk", "right") for coord in self.coordinates])
        conds = product(
            *[
                (
                    Eq(coord.idx, coord.idx.lower),
                    (coord.idx.lower < coord.idx) & (coord.idx < coord.idx.upper),
                    Eq(coord.idx, coord.idx.upper),
                )
                for coord in self.coordinates
            ]
        )
        return dict(zip(domains, [reduce(and_, cond) for cond in conds]))

    @property
    def node_coords(self):
        domains = product(*[("left", "bulk", "right") for coord in self.coordinates])

        return [
            tuple(
                coord.get_node_coord(domain)
                for coord, domain in zip(self.coordinates, domain)
            )
            for domain in domains
        ]

    @property
    def node_coords_subs(self):

        return [
            {
                coord.idx: node_coord
                for coord, node_coord in zip(self.coordinates, node_coords)
            }
            for node_coords in self.node_coords
        ]

    def is_in_bulk(self, *node_coords):
        return tuple(
            [
                coord.is_in_bulk(node_coord)
                for coord, node_coord in zip(self.coords, node_coords)
            ]
        )

    @property
    def symbol(self):
        return Function(self.name) if self.coords else Symbol(self.name)

    @property
    def discrete(self):
        return IndexedBase(self.name) if self.coords else Symbol(self.name)

    def __len__(self):
        return len(self.coords)

    @property
    def discrete_i(self):
        return self.discrete[self.c_idxs] if self.coords else Symbol(self.name)

    def __repr__(self):
        return "{}{}".format(
            self.name,
            ("(%s)" % ", ".join(map(str, self.coords)) if self.coords else ""),
        )

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return hash(self.__str__())


def _convert_unk_list(unknowns):
    def convert_unknown(unk):
        if isinstance(unk, str):
            return Unknown(unk)
        if isinstance(unk, Unknown):
            return unk
        raise ValueError("dependent var should be string or Unknown")

    if not isinstance(unknowns, (list, tuple)):
        return [convert_unknown(unknowns)]
    else:
        return tuple(list(map(convert_unknown, unknowns)))


def _convert_coord_list(coordinates):
    def convert_coord(coordinate):
        if isinstance(coordinate, str):
            return Coordinate(coordinate)
        if isinstance(coordinate, Coordinate):
            return coordinate
        raise ValueError("dependent var should be string or Coordinate")

    if not isinstance(coordinates, (list, tuple)):
        return [convert_coord(coordinates)]
    else:
        return tuple(list(map(convert_coord, coordinates)))



#!/usr/bin/env python
# coding=utf-8

from itertools import chain, product
from typing import Dict, Sequence

import attr
from sympy import Derivative, Max, Min

from .variables import Coordinate


def as_finite_diff(derivative, coord, offset=0, accuracy=2, scheme="centered"):
    try:
        order = dict(derivative.variable_count)[coord.symbol]
    except KeyError:
        return derivative
    if scheme == "centered":
        n = (order + 1) // 2 + accuracy - 2
        stencil = tuple(range(-n + offset, n + 1 + offset))
    elif scheme == "right":
        n = accuracy + order
        stencil = tuple(range(0 + offset, n + offset))

    elif scheme == "left":
        n = accuracy + order
        stencil = tuple(range(-(n - 1) + offset, 1 + offset))

    else:
        raise NotImplementedError(
            "scheme should be one of 'centered', 'left' or 'right'."
        )
    points = [coord.symbol + i * coord.step for i in stencil]
    return derivative.as_finite_difference(points=points, wrt=coord.symbol), stencil


def all_derivatives(derivative, wrt):
    return True


@attr.s()
class FiniteDifferenceScheme:
    scheme = attr.ib(type=str, default="centered")
    accuracy = attr.ib(type=int, default=2)
    offset = attr.ib(type=int, default=0)
    pattern = attr.ib(default=all_derivatives, repr=False)

    def relevant_derivatives(self, expr):
        derivatives = expr.atoms(Derivative)
        return list(
            [
                (deriv, coord)
                for deriv, coord in chain(
                    *[
                        product([derivative], derivative.variables)
                        for derivative in derivatives
                    ]
                )
                if self.pattern(deriv, coord)
            ]
        )

    def pop_derivative(self, expr):
        derivative, coord = list(self.relevant_derivatives(expr))[0]
        coord = Coordinate(str(coord))
        return derivative, coord

    def apply(self, expr):
        while self.relevant_derivatives(expr):
            derivative, coord = self.pop_derivative(expr)
            discrete_derivative, stencil = as_finite_diff(
                derivative,
                coord,
                accuracy=self.accuracy,
                offset=self.offset,
                scheme=self.scheme,
            )
            expr = expr.replace(derivative, discrete_derivative)
        return expr


def upwind(velocity, unk, coord=None, accuracy=2):
    if coord is None and len(unk.args) == 1:
        coord = unk.args[0]
    elif coord is None:
        raise ValueError(
            "You have to provide the coordinate where the upwind scheme apply for non 1D unknown."
        )

    def left_deriv(coord, unk):
        deriv = Derivative(unk, (coord, 1))
        n = accuracy + 1
        if accuracy < 3:
            stencil = tuple(range(-(n - 1), 1))

        elif accuracy == 3:
            stencil = tuple(range(-(n - 2), 2))
        else:
            raise NotImplementedError("Upwind is only available for n <= 3")
        points = [coord.symbol + i * coord.step for i in stencil]
        discretized_deriv = deriv.as_finite_difference(points=points, wrt=coord.symbol)
        return discretized_deriv

    def right_deriv(coord, unk):
        deriv = Derivative(unk, (coord, 1))
        n = accuracy + 1
        if accuracy < 3:
            stencil = tuple(range(0, n))
        elif accuracy == 3:
            stencil = tuple(range(-1, n - 1))
        else:
            raise NotImplementedError("Upwind is only available for n <= 3")
        points = [coord.symbol + i * coord.step for i in stencil]
        discretized_deriv = deriv.as_finite_difference(points=points, wrt=coord.symbol)
        return discretized_deriv

    ap = Max(velocity, 0)
    am = Min(velocity, 0)

    coord = Coordinate(str(coord))

    deriv_left = left_deriv(coord, unk)
    deriv_right = right_deriv(coord, unk)
    discretized_deriv = ap * deriv_left + am * deriv_right
    return discretized_deriv


def chain_schemes(schemes, expr, default_scheme="centered", default_accuracy=2):
    for scheme in [*schemes, FiniteDifferenceScheme(default_scheme, default_accuracy)]:
        expr = scheme.apply(expr)
    return expr
